__all__ = ()

import reprlib

from scarletio import Compound

from ...application import Application
from ...bases import maybe_snowflake
from ...builder.serialization import create_serializer
from ...builder.serialization_configuration import SerializationConfiguration
from ...channel import Channel, ChannelType, create_partial_channel_from_id
from ...http import DiscordApiClient, VALID_ICON_MEDIA_TYPES_EXTENDED
from ...message import Message
from ...message.message_builder import MessageBuilderWebhookCreate, MessageBuilderWebhookEdit
from ...utils import get_image_media_type, image_to_base64
from ...webhook import Webhook, create_partial_webhook_from_id

from ..request_helpers import (
    get_channel_id, get_guild_id, get_webhook_and_id, get_webhook_and_id_and_token, get_webhook_id,
    get_webhook_id_and_token,
)


MESSAGE_SERIALIZER_WEBHOOK_CREATE = create_serializer(
    MessageBuilderWebhookCreate,
    SerializationConfiguration(
        [
            MessageBuilderWebhookCreate.allowed_mentions,
            MessageBuilderWebhookCreate.attachments,
            MessageBuilderWebhookCreate.applied_tag_ids,
            MessageBuilderWebhookCreate.avatar_url,
            MessageBuilderWebhookCreate.components,
            MessageBuilderWebhookCreate.content,
            MessageBuilderWebhookCreate.embeds,
            MessageBuilderWebhookCreate.flags,
            MessageBuilderWebhookCreate.name,
            MessageBuilderWebhookCreate.poll,
            MessageBuilderWebhookCreate.thread_name,
            MessageBuilderWebhookCreate.tts,
            MessageBuilderWebhookCreate.voice_attachment,
        ],
        False,
    )
)

MESSAGE_SERIALIZER_WEBHOOK_EDIT = create_serializer(
    MessageBuilderWebhookEdit,
    SerializationConfiguration(
        [
            MessageBuilderWebhookEdit.allowed_mentions,
            MessageBuilderWebhookEdit.attachments,
            MessageBuilderWebhookEdit.components,
            MessageBuilderWebhookEdit.content,
            MessageBuilderWebhookEdit.embeds,
            MessageBuilderWebhookEdit.flags,
        ],
        True,
    )
)


class ClientCompoundWebhookEndpoints(Compound):
    
    application : Application
    api : DiscordApiClient
    
    
    async def webhook_create(self, channel, name, *, avatar = None):
        """
        Creates a webhook at the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel of the created webhook.
        name : `str`
            The name of the new webhook. It's length can be in range [1:80].
        avatar : `None`, `bytes-like` = `None`, Optional (Keyword only)
            The webhook's avatar. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. However if set as
            `'gif'`, it will not have any animation.
            
        Returns
        -------
        webhook : ``Webhook``
            The created webhook.
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel`` nor as `int`.
            - If `avatar` was not given neither as `None`, `bytes-like`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not given as `str`.
            - If `name` range is out of the expected range [1:80].
            - If `avatar`'s type is not any of the expected ones: `'jpg'`, `'png'`, `'webp'`, `'gif'`.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_system)
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {type(name).__name__}; {name!r}.'
                )
            
            name_length = len(name)
            if name_length < 1 or name_length > 80:
                raise AssertionError(
                    f'`name` length can be in range [1:80], got {name_length!r}; {name!r}.'
                )
        
        data = {'name': name}
        
        if (avatar is not None):
            if not isinstance(avatar, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`avatar` can `None`, `bytes-like`, got {avatar.__name__}; {reprlib.repr(avatar)}.'
                )
            
            if __debug__:
                media_type = get_image_media_type(avatar)
                if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                    raise AssertionError(
                        f'Invalid `avatar` media type: {media_type}; got {reprlib.repr(avatar)}.'
                    )
            
            data['avatar'] = image_to_base64(avatar)
        
        data = await self.api.webhook_create(channel_id, data)
        return Webhook.from_data(data)
    
    
    async def webhook_get(self, webhook):
        """
        Requests the webhook by it's id.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `int`
            The webhook to update or the webhook's id to get.
        
        Returns
        -------
        webhook : ``Webhook``
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.webhook_get_token`` : Getting webhook with Discord's webhook API.
        
        Notes
        -----
        If the webhook already loaded and if it's guild's webhooks are up to date, no request is done.
        """
        webhook, webhook_id = get_webhook_and_id(webhook)
        
        data = await self.api.webhook_get(webhook_id)
        if webhook is None:
            webhook = Webhook.from_data(data)
        else:
            webhook._update_attributes(data)
        
        return webhook
    
    
    async def webhook_get_token(self, webhook):
        """
        Requests the webhook through Discord's webhook API. The client do not needs to be in the guild of the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook to update or the webhook's id and token.
        
        Returns
        -------
        webhook : ``Webhook``
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the webhook already loaded and if it's guild's webhooks are up to date, no request is done.
        """
        webhook, webhook_id, webhook_token = get_webhook_and_id_and_token(webhook)
        
        if (webhook is None):
            webhook = create_partial_webhook_from_id(webhook_id, webhook_token)
        
        data = await self.api.webhook_get_token(webhook_id, webhook_token)
        webhook._set_attributes(data)
        return webhook
    
    
    async def webhook_get_all_channel(self, channel):
        """
        Requests all the webhooks of the channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's webhooks will be requested.
        
        Returns
        -------
        webhooks : `list` of ``Webhook` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
            
            You may expect the following exceptions:
            
            +---------------+-----------------------+---------------------------------------------------------------+
            | Error code    | Internal name         | Reason                                                        |
            +===============+=======================+===============================================================+
            | 10003         | unknown_channel       | The channel not exists.                                       |
            +---------------+-----------------------+---------------------------------------------------------------+
            | 50013         | missing_permissions   | You need `manage_webhooks` permission. (Or the client has no  |
            |               |                       | access to the channel.)                                       |
            +---------------+-----------------------+---------------------------------------------------------------+
            | 60003         | mfa_required          | You need to have multi-factor authorization to do this        |
            |               |                       | operation (guild setting dependent). For bot accounts it      |
            |               |                       | means their owner needs mfa.                                  |
            +---------------+-----------------------+---------------------------------------------------------------+
            
            > Discord drops `Forbidden (403), code = 50013: Missing Permissions` instead of
            > `Forbidden (403), code = 50001: Missing Access`.
        
        AssertionError
            If `channel` was given as a channel's identifier but it detectably not refers to a ``Channel``.
        
        Notes
        -----
        No request is done, if the passed channel is partial, or if the channel's guild's webhooks are up to date.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_system)
        
        data = await self.api.webhook_get_all_channel(channel_id)
        return [Webhook.from_data(webhook_data) for webhook_data in data]
    
    
    async def webhook_get_own_channel(self, channel):
        """
        Requests the webhooks of the given channel and returns the first owned one.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's webhooks will be requested.
        
        Returns
        -------
        webhooks : `list` of ``Webhook` objects
        """
        webhooks = await self.webhook_get_all_channel(channel)
        
        application_id = self.application.id
        for webhook in webhooks:
            if webhook.application_id == application_id:
                return webhook
        
        return None
    
    
    async def webhook_get_all_guild(self, guild):
        """
        Requests the webhooks of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``int | Guild``
            The guild, what's webhooks will be requested.
        
        Returns
        -------
        webhooks : `list` of ``Webhook` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        No request is done, if the guild's webhooks are up to date.
        """
        guild_id = get_guild_id(guild)
        
        webhook_datas = await self.api.webhook_get_all_guild(guild_id)
        return [Webhook.from_data(webhook_data) for webhook_data in webhook_datas]
    
    
    async def webhook_delete(self, webhook):
        """
        Deletes the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `int`
            The webhook to delete.
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.webhook_delete_token`` : Deleting webhook with Discord's webhook API.
        """
        webhook_id = get_webhook_id(webhook)
        
        await self.api.webhook_delete(webhook_id)
    
    
    async def webhook_delete_token(self, webhook):
        """
        Deletes the webhook through Discord's webhook API.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook to delete.

        Parameters
        ----------
        webhook : ``Webhook``
            The webhook to delete.
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        await self.api.webhook_delete_token(webhook_id, webhook_token)
    
    # later there gonna be more stuff that's why 2 different
    async def webhook_edit(self, webhook, *, name = ..., avatar = ..., channel = ...):
        """
        Edits and updates the given webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `int`
            The webhook to edit.
        name : `str`, Optional (Keyword only)
            The webhook's new name. It's length can be in range [1:80].
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            The webhook's new avatar. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. However if set as
            `'gif'`, it will not have any animation. If passed as `None`, will remove the webhook's current avatar.
        channel : ``Channel``, `int`, Optional (Keyword only)
            The webhook's channel.
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as `int`.
            - If `avatar` was not given neither as `None` nor as `bytes-like`.
            - If `channel` was not given neither as ``Channel`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was given but not as `str`.
            - If `name`'s length is out of range [1:80].
            - If `avatar`'s type is not any of the expected ones: `'jpg'`, `'png'`, `'webp'`, `'gif'`.
        
        See Also
        --------
        ``.webhook_edit_token`` : Editing webhook with Discord's webhook API.
        """
        webhook_id = get_webhook_id(webhook)
        
        data = {}
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {type(name).__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 1 or name_length > 80:
                    raise AssertionError(
                        f'The length of the name can be in range [1:80], got {name_length}; {name!r}.'
                    )
            
            data['name'] = name
        
        if (avatar is not ...):
            if avatar is None:
                avatar_data = None
            else:
                if not isinstance(avatar, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`avatar` can be `None`, `bytes-like`, got {type(avatar).__name__}; '
                        f'{reprlib.repr(avatar)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(avatar)
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid `avatar` type for the client: {media_type}; got {reprlib.repr(avatar)}.'
                        )
                
                avatar_data = image_to_base64(avatar)
            
            data['avatar'] = avatar_data
        
        if (channel is not ...):
            while True:
                if isinstance(channel, Channel):
                    if channel.is_in_group_guild_system() or channel.partial:
                        channel_id = channel.id
                        break
                
                else:
                    channel_id = maybe_snowflake(channel)
                    if channel_id is not None:
                        break
                
                raise TypeError(
                    f'`channel` can be gild text channel, `int`, got {type(channel).__name__}; '
                    f'{channel!r}.'
                )
            
            data['channel_id'] = channel_id
        
        if not data:
            return # Save 1 request
        
        data = await self.api.webhook_edit(webhook_id, data)
        webhook._set_attributes(data)
    
    
    async def webhook_edit_token(self, webhook, *, name = ..., avatar = ...):
        """
        Edits and updates the given webhook through Discord's webhook API.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook to edit.
        name : `str`, Optional (Keyword only)
            The webhook's new name. It's length can be between `1` and `80`.
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            The webhook's new avatar. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. However if set as
            `'gif'`, it will not have any animation. If passed as `None`, will remove the webhook's current avatar.
        
        Returns
        -------
        webhook : ``Webhook``
            The updated webhook.
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
            - If `avatar` was not given neither as `None` nor as `bytes-like`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was given but not as `str`.
            - If `name`'s length is out of range [1:80].
            - If `avatar`'s type is not any of the expected ones: `'jpg'`, `'png'`, `'webp'`, `'gif'`.
        
        Notes
        -----
        This endpoint cannot edit the webhook's channel, like ``.webhook_edit``.
        """
        webhook, webhook_id, webhook_token = get_webhook_and_id_and_token(webhook)
        
        data = {}
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `None`, `str`, got {type(name).__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 1 or name_length > 80:
                    raise AssertionError(
                        f'`name`\'s length can be in range [1:80], got {name_length}; {name!r}.'
                    )
            
            data['name'] = name
        
        if (avatar is not ...):
            if avatar is None:
                avatar_data = None
            else:
                if not isinstance(avatar, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`avatar` can be `None`, `bytes-like`, got {type(avatar).__name__}; '
                        f'{reprlib.repr(avatar)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(avatar)
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid avatar type for the client: {media_type}; got {reprlib.repr(avatar)}.'
                        )
                
                avatar_data = image_to_base64(avatar)
            
            data['avatar'] = avatar_data
        
        if not data:
            return # Save 1 request
        
        data = await self.api.webhook_edit_token(webhook_id, webhook_token, data)
        
        if webhook is None:
            webhook = Webhook.from_data(data)
        else:
            webhook._set_attributes(data)
        
        return webhook
    
    
    async def webhook_message_create(
        self, webhook, *positional_parameters, thread = None, wait = False, **keyword_parameters,
    ):
        """
        Sends a message with the given webhook. If there is nothing to send, or if `wait` was not passed as `True`
        returns `None`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook through what will the message be sent.
        
        *positional_parameters : Positional parameters
            Additional parameters to create the message with.
        
        thread : ``None | int | Channel`` = `None`, Optional (Keyword only)
            The thread of the webhook's channel where the message should be sent.
        
        wait : `None`, `bool` = `None`, Optional (Keyword only)
            Whether we should wait for the message to send and receive it's data as well.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to create the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        applied_tag_ids : `None`, `(list | tuple)<int | ForumTag>`, `int`, `ForumTag`, Optional (Keyword only)
             The tags' identifier which have been applied to the thread.
             Applicable for threads of a forum-like channels.
        
        applied_tags : `None`, `(list | tuple)<int | ForumTag>`, `int`, `ForumTag`, Optional (Keyword only)
            Alternative for `applied_tag_ids`.
        
        attachments : `None | object`, Optional (Keyword only)
            Attachments to send.
        
        avatar_url : `None`, `str` = `None`, Optional (Keyword only)
            The message's author's avatar's url. Defaults to the webhook's avatar' url by Discord.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
        
        content : `None`, `str`, Optional
            The message's content if given.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
        
        file : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's flags.
        
        name : `None`, `str` = `None`, Optional (Keyword only)
            The message's author's new name. Default to the webhook's name by Discord.
        
        poll : `None`, ``Poll``, Optional
            The message's poll.
        
        silent : `bool` = `False`, Optional (Keyword only)
            Whether the message should be delivered silently.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        thread_name : `None`, `str`, Optional (Keyword only)
            The thread's name to create. Applicable only in forum-like channels.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        voice_attachment : ``None | VoiceAttachment``, Optional (Keyword only)
            Modifies the message to be a voice message, allowing it to contain just a single voice attachment.
        
        Returns
        -------
        message : ``Message``, `None`
            Returns `None` if there is nothing to send or if `wait` was given as `False` (so by default).
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        - ``.message_create`` : Create a message with your client.
        - ``.webhook_message_edit`` : Edit a message created by a webhook.
        - ``.webhook_message_delete`` : Delete a message created by a webhook.
        - ``.webhook_message_get`` : Get a message created by a webhook.
        """
        webhook, webhook_id, webhook_token = get_webhook_and_id_and_token(webhook)
        
        while True:
            if thread is None:
                thread_id = 0
                break
                
            elif isinstance(thread, Channel):
                if thread.is_in_group_thread() or thread.partial:
                    thread_id = thread.id
                    break
            
            else:
                thread_id = maybe_snowflake(thread)
                if thread_id is not None:
                    break
            
            raise TypeError(
                f'`thread` can be `None`, thread channel, `int`, '
                f'got {type(thread).__name__}; {thread!r}.'
            )
        
        if not isinstance(wait, bool):
            raise TypeError(
                f'`wait` can be `bool`, got {type(wait).__name__}; {wait!r}.'
            )
        
        message_data = MESSAGE_SERIALIZER_WEBHOOK_CREATE(positional_parameters, keyword_parameters)
        if not message_data:
            return
        
        query = None
        if wait:
            if query is None:
                query = {}
            
            query['wait'] = wait
        
        if thread_id:
            if query is None:
                query = {}
            
            query['thread_id'] = str(thread_id)
        
        if (webhook is None) or (not webhook.application_id):
            if query is None:
                query = {}
            
            query['with_components'] = True
        
        data = await self.api.webhook_message_create(webhook_id, webhook_token, message_data, query)
        
        if not wait:
            return
        
        # Use goto
        while True:
            # Created into thread
            if thread_id:
                if thread is not None:
                    channel = thread
                    break
                
                channel = create_partial_channel_from_id(thread_id, ChannelType.guild_thread_public, 0)
                break
            
            # Created as thread
            try:
                thread_data = data['thread']
            except KeyError:
                pass
            else:
                channel = create_partial_channel_from_id(
                    int(thread_data['id']),
                    ChannelType.guild_thread_public,
                    0,
                )
                break
            
            # Check webhook.channel
            if (webhook is not None):
                channel = webhook.channel
                if (channel is not None):
                    break
            
            # Create new channel if unknown.
            # Create thread if `thread_name` is present.
            channel = create_partial_channel_from_id(
                int(data['channel_id']),
                ChannelType.guild_text,
                0,
            )
            break
        
        return channel._create_new_message(data)
    
    
    async def webhook_message_edit(self, webhook, message, *positional_parameters, **keyword_parameters):
        """
        Edits the message sent by the given webhook. The message's author must be the webhook itself.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook who created the message.
        
        message : ``Message``, `int`
            The webhook's message to edit.
        
        *positional_parameters : Positional parameters
            Additional parameters to edit the message with.
        
        **keyword_parameters : Keyword parameters
            Additional parameters to edit the message with.
        
        Other Parameters
        ----------------
        allowed_mentions : `None`,  ``AllowedMentionProxy``, `str`, ``UserBase``, ``Role``, `list` of \
                (`str`, ``UserBase``, ``Role`` ) , Optional
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        attachments : `None | object`, Optional (Keyword only)
            Attachments to send.
        
        components : `None`, ``Component``, `(tuple | list)<Component, (tuple | list)<Component>>`
            Components attached to the message.
            
            Pass it as `None` remove the actual ones.
        
        content : `None`, `str`, Optional
            The new content of the message.
        
        embed : `None`, `Embed`, Optional
            Alternative for `embeds`.
        
        embeds : `None`, `list<Embed>`, Optional
            The new embedded content of the message.
            
            By passing it as `None`, you can remove the old.
        
        file : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        files : `None | object`, Optional (Keyword only)
            Alternative for `attachments`.
        
        flags : `int`, ``MessageFlag`, Optional
            The message's new flags.
        
        suppress_embeds : `bool`, Optional (Keyword only)
            Whether the message's embeds should be suppressed or unsuppressed.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        - ``.message_edit`` : Edit your own messages.
        - ``.webhook_message_create`` : Create a message with a webhook.
        - ``.webhook_message_delete`` : Delete a message created by a webhook.
        - ``.webhook_message_get`` : Get a message created by a webhook.
        
        Notes
        -----
        Embed messages ignore suppression with their endpoint, not like ``.message_edit`` endpoint.
        
        Editing the message with empty string is broken.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        # Detect message id
        # 1.: Message
        # 2.: int (str)
        # 4.: None -> raise
        # 5.: raise
        
        if isinstance(message, Message):
            if message.author.id != webhook_id:
                raise RuntimeError(
                    f'The message was not sent by the webhook, got {message!r}; '
                    f'webhook_id = {webhook_id!r}; message.author.id = {message.author.id!r}.'
                )
                
            message_id = message.id
        
        else:
            message_id = maybe_snowflake(message)
            if (message_id is not None):
                pass
            
            elif message is None:
                raise TypeError(
                    f'`message` was given as `None`. Make sure to use '
                    f'`{type(self).__name__}.webhook_message_create` with giving content and by passing '
                    f'`wait` parameter as `True` as well.'
                )
            
            else:
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, int`, got '
                    f'{type(message).__name__}; {message!r}.'
                )
        
        message_data = MESSAGE_SERIALIZER_WEBHOOK_EDIT(positional_parameters, keyword_parameters)
        if not message_data:
            return
        
        query = None
        if (webhook is None) or (not webhook.application_id):
            if query is None:
                query = {}
            
            query['with_components'] = True
        
        # We receive the new message data, but we do not update the message, so dispatch events can get the difference.
        await self.api.webhook_message_edit(webhook_id, webhook_token, message_id, message_data, query)
    
    
    async def webhook_message_delete(self, webhook, message):
        """
        Deletes the message sent by the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook who created the message.
        message : ``Message``, `int`
            The webhook's message to delete.
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `int`.
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `message` was detectably not sent by the `webhook`.
        
        See Also
        --------
        - ``.message_delete`` : Delete a message.
        - ``.webhook_message_create`` : Create a message with a webhook.
        - ``.webhook_message_edit`` : Edit a message created by a webhook.
        - ``.webhook_message_get`` : Get a message created by a webhook.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        # Detect message id
        # 1.: Message
        # 2.: int
        # 4.: None -> raise
        # 5.: raise
        
        if isinstance(message, Message):
            if __debug__:
                if message.author.id != webhook_id:
                    raise AssertionError(
                        f'The message was not sent by the webhook, got {message!r}.'
                    )
            
            message_id = message.id
        
        else:
            message_id = maybe_snowflake(message)
            if (message_id is not None):
                pass
            
            elif message is None:
                raise TypeError(
                    f'`message` was given as `None`. Make sure to use '
                    f'`{type(self).__name__}.webhook_message_create` with giving content and by passing '
                    f'`wait` parameter as `True` as well.'
                )
            
            else:
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, `int`, got '
                    f'{type(message).__name__}; {message!r}.'
                )
        
        await self.api.webhook_message_delete(webhook_id, webhook_token, message_id)
    
    
    async def webhook_message_get(self, webhook, message_id):
        """
        Gets a previously sent message with the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook who created the message.
        message_id : `int`
            The webhook's message's identifier to get.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
            - If `message_id` was not given as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        - ``.message_get`` : Get a message.
        - ``.webhook_message_create`` : Create a message with a webhook.
        - ``.webhook_message_edit`` : Edit a message created by a webhook.
        - ``.webhook_message_delete`` : Delete a message created by a webhook.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        message_id_value = maybe_snowflake(message_id)
        if message_id_value is None:
            raise TypeError(
                f'`message_id` can be `int`, got {type(message_id).__name__}; {message_id!r}.'
            )
        
        message_data = await self.api.webhook_message_get(webhook_id, webhook_token, message_id_value)
        return Message.from_data(message_data)



from .soundboard_sounds_event_handler import *
from .soundboard_sounds_event_waiter import *


__all__ = (
    *soundboard_sounds_event_handler.__all__,
    *soundboard_sounds_event_waiter.__all__,
)



from .audio_settings import *
from .encryption_adapters import *
from .packets import *

from .audio_source import *
from .opus import *
from .player import *
from .reader import *
from .utils import *
from .voice_client import *


__all__ = (
    *audio_settings.__all__,
    *encryption_adapters.__all__,
    *packets.__all__,
    
    *audio_source.__all__,
    *opus.__all__,
    *player.__all__,
    *reader.__all__,
    *utils.__all__,
    *voice_client.__all__,
)



__all__ = ('MediaInfo', )

from scarletio import RichAttributeErrorBaseType

from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_attachment_id, parse_content_type, parse_height, parse_proxy_url, parse_url, parse_width, put_attachment_id,
    put_content_type, put_height, put_proxy_url, put_url, put_width, validate_attachment_id, validate_content_type,
    validate_height, validate_proxy_url, validate_url, validate_width
)


PRECREATE_FIELDS = {
    'attachment_id': ('attachment_id', validate_attachment_id),
    'content_type': ('content_type', validate_content_type),
    'height': ('height', validate_height),
    'proxy_url': ('proxy_url', validate_proxy_url),
    'url': ('url', validate_url),
    'width': ('width', validate_width),
}


class MediaInfo(RichAttributeErrorBaseType):
    """
    Represents a media information of a ``MediaItem`` or of a few components that may support embedding it.
    
    Attributes
    ----------
    attachment_id : `int`
        The identifier of the represented attachment if any.
    
    content_type : `None | str`
        The media info's media type.
    
    height : `int`
        The height of the media info if applicable.
        
        > Defaults to `0`.
    
    proxy_url : `str`
        Proxied url of the media info.
    
    url : `str`
        The media info's url.
    
    width : `int`
        The media info's width if applicable.
        
        > Defaults to `0`.
    """
    __slots__ = ('attachment_id', 'content_type', 'height', 'proxy_url', 'url', 'width')
    
    
    def __new__(cls, url):
        """
        Creates a new media info.
        
        Parameters
        ----------
        url : `str`.
            The media's url.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        url = validate_url(url)
        
        # Construct
        self = object.__new__(cls)
        self.attachment_id = 0
        self.content_type = None
        self.height = 0
        self.proxy_url = None
        self.url = url
        self.width = 0
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an empty media info.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.attachment_id = 0
        self.content_type = None
        self.height = 0
        self.proxy_url = None
        self.url = ''
        self.width = 0
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a media info from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data to create form.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.attachment_id = parse_attachment_id(data)
        self.content_type = parse_content_type(data)
        self.height = parse_height(data)
        self.proxy_url = parse_proxy_url(data)
        self.url = parse_url(data)
        self.width = parse_width(data)
        return self
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__,]
        
        # url
        repr_parts.append(' url = ')
        repr_parts.append(repr(self.url))
        
        # Extra if image
        width = self.width
        height = self.height
        if width and height:
            repr_parts.append(',  = ')
            repr_parts.append(repr(width))
            repr_parts.append('x')
            repr_parts.append(repr(height))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_proxy_url = self.proxy_url
        other_proxy_url = other.proxy_url
        if (self_proxy_url is not None) and (other_proxy_url is not None):
            # proxy_url
            if self.proxy_url != other.proxy_url:
                return False
            
            # attachment_id
            if self.attachment_id != other.attachment_id:
                return False
            
            # content_type
            if self.content_type != other.content_type:
                return False
            
            # height
            if self.height != other.height:
                return False
            
            # width
            if self.width != other.width:
                return False
        
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # proxy_url
        proxy_url = self.proxy_url
        if (proxy_url is not None):
            hash_value ^= hash(proxy_url)
            
            # attachment_id
            hash_value ^= self.attachment_id
            
            # content_type
            content_type = self.content_type
            if (content_type is not None):
                hash_value ^= hash(content_type)
            
            # height
            hash_value ^= self.height << 16
            
            # width
            hash_value ^= self.width << 8
        
        # url
        hash_value ^= hash(self.url)
        
        return hash_value
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Serializes the media info.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_url(self.url, data, defaults)
        
        if include_internals:
            put_attachment_id(self.attachment_id, data, defaults)
            put_content_type(self.content_type, data, defaults)
            put_height(self.height, data, defaults)
            put_width(self.width, data, defaults)
            put_proxy_url(self.proxy_url, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the media info.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.attachment_id = 0
        new.content_type = None
        new.height = 0
        new.proxy_url = None
        new.url = self.url
        new.width = 0
        return new
    
    
    def copy_with(self, *, url = ...):
        """
        Copies the media info with the given fields.
        
        Parameters
        ----------
        url : `str`, Optional (Keyword only)
            The media's url.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # url
        if url is ...:
            url = self.url
        else:
            url = validate_url(url)
        
        # Construct
        new = object.__new__(type(self))
        new.attachment_id = 0
        new.content_type = None
        new.height = 0
        new.proxy_url = None
        new.url = url
        new.width = 0
        return new
    
    
    @classmethod
    def precreate(
        cls,
        **keyword_parameters,
    ):
        """
        Precreates the media info. The only difference between ``.__new__`` and ``.precreate`` here is that
        ``.precreate`` allows passing internal fields as well.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        attachment_id : ``None | int | Attachment``, Optional (Keyword only)
            The identifier of the represented attachment if any.
        
        content_type : `None | str`, Optional (Keyword only)
            The media info's media type.
        
        height : `int`, Optional (Keyword only)
            The height of the media info if applicable.
        
        proxy_url : `None | str`, Optional (Keyword only)
            Proxied url of the media info.
        
        url : `str`, Optional (Keyword only)
            The media info's url.
        
        width : `int`, Optional (Keyword only)
            The media info's width if applicable.
        
        Returns
        -------
        self : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        
        else:
            processed = None
        
        # Construct
        self = cls._create_empty()
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self



__all__ = ()

from functools import partial as partial_func

from scarletio import RichAttributeErrorBaseType, include

from ....discord.events.handling_helpers import Router

from ..constants import APPLICATION_COMMAND_FUNCTION_DEEPNESS

from .command import CommandInterface


SlashCommandParameterAutoCompleter = include('SlashCommandParameterAutoCompleter')


def validate_auto_complete_parameter_name(parameter_name):
    """
    Checks out one parameter name to auto complete.
    
    Parameters
    ----------
    parameter_name : `str`
        The parameter's name to auto complete.
    
    Returns
    -------
    parameter_name : `str`
        The validated parameter name to autocomplete.
    
    Raises
    ------
    TypeError
        If `parameter_name` is not `str`.
    ValueError
        If `parameter_name` is an empty string.
    """
    if type(parameter_name) is str:
        pass
    
    elif isinstance(parameter_name, str):
        parameter_name = str(parameter_name)
    
    else:
        raise TypeError(
            f'`parameter_name` can be `str`, got '
            f'{parameter_name.__class__.__name__}; {parameter_name!r}.'
        )
    
    if not parameter_name:
        raise ValueError(
            f'`parameter_name` cannot be empty string.'
        )
    
    return parameter_name


def build_auto_complete_parameter_names(parameter_name, parameter_names):
    """
    Builds a checks out parameter names.
    
    Parameters
    ----------
    parameter_name : `str`
        The parameter's name to auto complete.
    parameter_names : `tuple<str>`
        Additional parameter to autocomplete.
    
    Returns
    -------
    processed_parameter_names : `list<str>`
        The processed parameter names.
    
    Raises
    ------
    TypeError
        If `parameter_name` is not `str`.
    ValueError
        If `parameter_name` is an empty string.
    """
    processed_parameter_names = []
    
    parameter_name = validate_auto_complete_parameter_name(parameter_name)
    processed_parameter_names.append(parameter_name)
    
    if parameter_names:
        for iterated_parameter_name in parameter_names:
            iterated_parameter_name = validate_auto_complete_parameter_name(iterated_parameter_name)
            if iterated_parameter_name not in processed_parameter_names:
                processed_parameter_names.append(iterated_parameter_name)
    
    return processed_parameter_names


def _register_auto_completer(parent, parameter_names, function):
    """
    Returned by `.autocomplete` decorators wrapped inside of `functools.partial` if `function` is not given.
    
    Parameters
    ----------
    parent : ``AutocompleteInterface``
        The parent entity to register the auto completer to.
    parameter_names : `list` of `str`
        The parameters' names.
    function : `async-callable`
        The function to register as auto completer.
    
    Returns
    -------
    auto_completer : `SlashCommandParameterAutoCompleter`
        The registered auto completer
    
    Raises
    ------
    RuntimeError
        - `function` cannot be `None`.
        - If the application command function has no parameter named, like `parameter_name`.
        - If the parameter cannot be auto completed.
    TypeError
        If `function` is not an asynchronous.
    """
    return parent._register_auto_completer(function, parameter_names)


class AutocompleteInterface(RichAttributeErrorBaseType):
    """
    Common class for auto completable objects.
    """
    __slots__ = ()
    
    def autocomplete(self, parameter_name, *parameter_names, function = ...):
        """
        Registers an auto completer function to the application command.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        
        *parameter_names : `str`
            Additional parameter names to autocomplete
        
        function : `None | async-callable`, Optional (Keyword only)
            The function to register as auto completer.
        
        Returns
        -------
        function / wrapper : `callable`, `functools.partial`
            The registered function if given or a wrapper to register the function with.
        
        Raises
        ------
        RuntimeError
            - If the parameter already has a auto completer defined.
            - If the application command function has no parameter named, like `parameter_name`.
            - If the parameter cannot be auto completed.
        TypeError
            If `function` is not an asynchronous.
        """
        parameter_names = build_auto_complete_parameter_names(parameter_name, parameter_names)
        
        if (function is ...):
            return partial_func(_register_auto_completer, self, parameter_names)
        
        return self._register_auto_completer(function, parameter_names)
    
    
    def _register_auto_completer(self, function, parameter_names):
        """
        Registers an autocomplete function.
        
        Parameters
        ----------
        function : `async-callable`
            The function to register as auto completer.
        
        parameter_names : `list<str>`
            The parameters' names.
        
        Returns
        -------
        auto_completer : `SlashCommandParameterAutoCompleter`
            The registered auto completer
        
        Raises
        ------
        RuntimeError
            - If the application command function has no parameter named, like `parameter_name`.
            - If the parameter cannot be auto completed.
        TypeError
            - If `function` is not an asynchronous.
        """
        auto_completer = self._make_auto_completer(function, parameter_names)
        self._store_auto_completer(auto_completer)
        return auto_completer
    
    
    def _store_auto_completer(self, auto_completer):
        """
        Stores an auto completer.
        
        Parameters
        -------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer to store.
        
        Returns
        -------
        stored : `bool`
        instance : ``SlashCommandParameterAutoCompleter``
        """
        auto_completers = self._auto_completers
        if (auto_completers is None):
            auto_completers = []
            self._auto_completers = auto_completers
        
        auto_completers.append(auto_completer)
        return True, auto_completer
    
    
    def _make_auto_completer(self, function, parameter_names):
        """
        Creates an auto completer.
        
        Parameters
        ----------
        function : `async-callable`
            The function to register as auto completer.
        parameter_names : `list<str>`
            The parameters' names.
        
        Returns
        -------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The created auto completer.
        
        Raises
        ------
        RuntimeError
            - If the application command function has no parameter named, like `parameter_name`.
            - If the parameter cannot be auto completed.
        TypeError
            - If `function` is not an asynchronous.
        """
        if isinstance(function, Router):
            function = function[0]
        
        if isinstance(function, CommandInterface):
            function = function.get_command_function()
        
        return SlashCommandParameterAutoCompleter(
            function,
            parameter_names,
            APPLICATION_COMMAND_FUNCTION_DEEPNESS,
            self,
        )
    
    
    @property
    def _auto_completers(self):
        """
        The registered auto completers.
        
        Overwrite it as an attribute in subclasses.
        
        Returns
        -------
        auto_completers : `None | list<SlashCommandParameterAutoCompleter>`
        """
        return None
    
    
    @_auto_completers.setter
    def _auto_completers(self, value):
        pass



__all__ = ('ScheduledEventOccasionOverwrite',)

from scarletio import RichAttributeErrorBaseType

from ...utils import DISCORD_EPOCH_START

from .fields import (
    parse_cancelled, parse_end, parse_start, parse_timestamp, put_cancelled, put_end, put_start, put_timestamp,
    validate_cancelled, validate_end, validate_start, validate_timestamp
)


class ScheduledEventOccasionOverwrite(RichAttributeErrorBaseType):
    """
    Represents a scheduled event occasion overwrite.
    
    Attributes
    ----------
    cancelled : `bool`
        Whether the occasion is cancelled.
    
    end : `None | DateTime`
        New end of the occasion.
    
    start : `None | DateTime`
        New start of the occasion.
    
    timestamp : `DateTime`
        The affected occasion.
    """
    __slots__ = ('cancelled', 'end', 'start', 'timestamp', )
    
    
    def __new__(cls, *, cancelled = ..., end = ..., start = ..., timestamp = ...):
        """
        Creates a new scheduled event occasion overwrite.
        
        Parameters
        ----------
        cancelled : `None | bool``, Optional (Keyword only)
            Whether the occasion is cancelled.
        
        end : `None | DateTime``, Optional (Keyword only)
            New end of the occasion.
        
        start : `None | DateTime``, Optional (Keyword only)
            New start of the occasion.
        
        timestamp : `DateTime`, Optional (Keyword only)
            The affected occasion.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # cancelled
        if cancelled is ...:
            cancelled = False
        else:
            cancelled = validate_cancelled(cancelled)
        
        # end
        if end is ...:
            end = None
        else:
            end = validate_end(end)
        
        # start
        if start is ...:
            start = None
        else:
            start = validate_start(start)
        
        # timestamp
        if timestamp is ...:
            timestamp = DISCORD_EPOCH_START
        else:
            timestamp = validate_timestamp(timestamp)
        
        # Construct
        self = object.__new__(cls)
        self.cancelled = cancelled
        self.end = end
        self.start = start
        self.timestamp = timestamp
        return self
    
    
    @classmethod
    def create_empty(cls):
        """
        Creates an empty occasion overwrite.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.cancelled = False
        self.end = None
        self.start = None
        self.timestamp = DISCORD_EPOCH_START
        return self
    
    
    @classmethod
    def from_fields(cls, timestamp, cancelled, start, end):
        """
        Creates a scheduled event occasion overwrite from the given already parsed fields.
        
        Parameters
        ----------
        timestamp : `DateTime`
            The affected occasion.
        
        cancelled : `bool`
            Whether the occasion is cancelled.
        
        start : `None | DateTime`
            New start of the occasion.
        
        end : `None | DateTime`
            New end of the occasion.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.timestamp = timestamp
        self.cancelled = cancelled
        self.start = start
        self.end = end
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new scheduled event occasion overwrite from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Scheduled event cancellation event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.cancelled = parse_cancelled(data)
        self.end = parse_end(data)
        self.start = parse_start(data)
        self.timestamp = parse_timestamp(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Serializes the scheduled event occasion overwrite.
        
        Parameters
        ----------
        defaults : `bool`
            Whether fields of their default value should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_cancelled(self.cancelled, data, defaults)
        put_end(self.end, data, defaults)
        put_start(self.start, data, defaults)
        
        if include_internals:
            put_timestamp(self.timestamp, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # timestamp
        repr_parts.append(' timestamp = ')
        repr_parts.append(repr(self.timestamp))
        
        # cancelled
        cancelled = self.cancelled
        if cancelled:
            repr_parts.append(', cancelled = ')
            repr_parts.append(repr(cancelled))
        
        # start
        start = self.start
        if (start is not None):
            repr_parts.append(', start = ')
            repr_parts.append(repr(start))
        
        # end
        end = self.end
        if (end is not None):
            repr_parts.append(', end = ')
            repr_parts.append(repr(end))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        # cancelled
        if self.cancelled != other.cancelled:
            return False
        
        # end
        if self.end != other.end:
            return False
        
        # start
        if self.start != other.start:
            return False
        
        # timestamp
        if self.timestamp != other.timestamp:
            return False
        
        return True
    
    
    def __gt__(self, other):
        """Returns self > other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.timestamp > other.timestamp
    
    
    def __lt__(self, other):
        """Returns self < other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.timestamp < other.timestamp
    
    
    def __hash__(self):
        """Returns hash(self)."""
        hash_value = 0
        
        # cancelled
        hash_value ^= self.cancelled
        
        # end
        end = self.end
        if (end is not None):
            hash_value ^= hash(end)
        
        # start
        start = self.start
        if (start is not None):
            hash_value ^= hash(start)
        
        # timestamp
        hash_value ^= hash(self.timestamp)
        
        return hash_value
    

    def copy(self):
        """
        Copies the scheduled event occasion overwrite.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.cancelled = self.cancelled
        new.end = self.end
        new.start = self.start
        new.timestamp = self.timestamp
        return new
    
    
    def copy_with(self, *, cancelled = ..., end = ..., start = ..., timestamp = ...):
        """
        Copies the scheduled event occasion overwrite with the given fields.
        
        Parameters
        ----------
        cancelled : `None | bool``, Optional (Keyword only)
            Whether the occasion is cancelled.
        
        end : `None | DateTime``, Optional (Keyword only)
            New end of the occasion.
        
        start : `None | DateTime``, Optional (Keyword only)
            New start of the occasion.
        
        timestamp : `DateTime`, Optional (Keyword only)
            The affected occasion.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # cancelled
        if cancelled is ...:
            cancelled = self.cancelled
        else:
            cancelled = validate_cancelled(cancelled)
        
        # end
        if end is ...:
            end = self.end
        else:
            end = validate_end(end)
        
        # start
        if start is ...:
            start = self.start
        else:
            start = validate_start(start)
        
        # timestamp
        if timestamp is ...:
            timestamp = self.timestamp
        else:
            timestamp = validate_timestamp(timestamp)
        
        # Construct
        new = object.__new__(type(self))
        new.cancelled = cancelled
        new.end = end
        new.start = start
        new.timestamp = timestamp
        return new
    
    
    def _update_attributes(self, data):
        """
        Updates the attributes of the scheduled event occasion overwrite.
        
        Parameters
        ----------
        data : `dict<str, object>`
            State data given.
        """
        self.cancelled = parse_cancelled(data)
        self.end = parse_end(data)
        self.start = parse_start(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the attributes of the scheduled event and returns the changed ones within an `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Cancellation data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            The updated attributes.
            
            The returned might have the following items:
            
            +-----------+-------------------+
            | Key       | Value             |
            +===========+===================+
            | cancelled | `bool`            |
            +-----------+-------------------+
            | end       | `None | DateTime` |
            +-----------+-------------------+
            | start     | `None | DateTime` |
            +-----------+-------------------+
        """
        old_attributes = {}
        
        # cancelled
        cancelled = parse_cancelled(data)
        if self.cancelled != cancelled:
            old_attributes['cancelled'] = self.cancelled
            self.cancelled = cancelled
        
        # end
        end = parse_end(data)
        if self.end != end:
            old_attributes['end'] = self.end
            self.end = end
        
        # start
        start = parse_start(data)
        if self.start != start:
            old_attributes['start'] = self.start
            self.start = start
        
        return old_attributes



__all__ = ('GuildBoost',)


from ...bases import DiscordEntity
from ...core import GUILDS, GUILD_BOOSTS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import create_partial_user_from_id

from .fields import (
    parse_guild_id, parse_user_id, put_guild_id, put_user_id, validate_guild_id, validate_user_id,
    parse_id, put_id, validate_id,
    parse_ended, put_ended, validate_ended,
    parse_ends_at, put_ends_at, validate_ends_at,
    parse_paused_until, validate_paused_until, put_paused_until
)

PRECREATE_FIELDS = {
    'ended': ('ended', validate_ended),
    'ends_at': ('ends_at', validate_ends_at),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'paused_until': ('paused_until', validate_paused_until),
    'user': ('user_id', validate_user_id),
    'user_id': ('user_id', validate_user_id),
}



class GuildBoost(DiscordEntity, immortal = True):
    """
    Represents a guild's boost.
    
    Attributes
    ----------
    ended : `bool`
        Whether the boost already ended.
    
    ends_at : `None | DateTime`
        When the boost ends.
    
    guild_id : `int`
        The guild's identifier the boost is for.
    
    id : `int`
        The guild boost' identifier.
    
    paused_until : `None | Datetime`
        Until when the boost is paused.
    
    user_id : `int`
        The user who owns the boost.
    """
    __slots__ = ('ended', 'ends_at', 'guild_id', 'paused_until', 'user_id', )
    
    def __new__(cls, *, paused_until = ...):
        """
        Creates a new guild boost.
        
        Parameters
        ----------
        paused_until : `None | DateTime`, Optional (Keyword only)
            Until when the boost is paused.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # paused_until
        if paused_until is ...:
            paused_until = None
        else:
            paused_until = validate_paused_until(paused_until)
        
        # Construct
        self = object.__new__(cls)
        self.ended = False
        self.ends_at = None
        self.guild_id = 0
        self.id = 0
        self.paused_until = paused_until
        self.user_id = 0
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild boost from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild boost data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        guild_boost_id = parse_id(data)
        
        try:
            self = GUILD_BOOSTS[guild_boost_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = guild_boost_id
            GUILD_BOOSTS[guild_boost_id] = self
        
        self._set_attributes(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Serializes the guild boost.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_paused_until(self.paused_until, data, defaults)
        
        if include_internals:
            put_ended(self.ended, data, defaults)
            put_ends_at(self.ends_at, data, defaults)
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
            put_user_id(self.user_id, data, defaults)
        
        return data
    
    
    def _set_attributes(self, data):
        """
        Sets the attributes of the guild boost from the given data (except id).
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild boost data.
        """
        self.guild_id = parse_guild_id(data)
        self.user_id = parse_user_id(data)
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the modifiable attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild boost data.
        """
        self.ended = parse_ended(data)
        self.ends_at = parse_ends_at(data)
        self.paused_until = parse_paused_until(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the modifiable attributes returns the changed attributes in a dictionary with the changed attributes
        in a `attribute-name`, `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild boost data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dict is optional.
            
            Might contain the following items:
            
            +-----------+-------------------+
            | Keys      | Values            |
            +===========+===================+
            | ended     | `bool`            |
            +-----------+-------------------+
            | ends_at   | `None | DateTime` |
            +-----------+-------------------+
        """
        old_attributes = {}
        
        # ended
        ended = parse_ended(data)
        if self.ended != ended:
            old_attributes['ended'] = self.ended
            self.ended = ended
        
        # ends_at
        ends_at = parse_ends_at(data)
        if self.ends_at != ends_at:
            old_attributes['ends_at'] = self.ends_at
            self.ends_at = ends_at
        
        # paused_until
        paused_until = parse_paused_until(data)
        if self.paused_until != paused_until:
            old_attributes['paused_until'] = self.paused_until
            self.paused_until = paused_until
        
        return old_attributes
    
    
    def __repr__(self):
        """Returns repr(self)."""
        repr_parts = ['<', type(self).__name__]
        
        # guild_id
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        # user_id
        repr_parts.append(', user_id = ')
        repr_parts.append(repr(self.user_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns hash(self)"""
        guild_boost_id = self.id
        if guild_boost_id:
            return guild_boost_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the guild boosts's hash based on their fields.
        
        This method is called by ``.__hash__`` if the guild boost has no ``.id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # ended -> skip, internal
        # ends_at -> skip, internal
        # guild_id -> skip, internal
        
        # paused_until
        paused_until = self.paused_until
        if (paused_until is not None):
            hash_value ^= hash(paused_until)
        
        # user_id -> skip, internal
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns self == other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns self != other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return (self_id == other_id)
        
        # ended -> skip, internal
        # ends_at -> skip, internal
        # guild_id -> skip, internal
        
        # paused_until
        if self.paused_until != other.paused_until:
            return False
        
        # user_id -> skip, internal
        
        return True
    
    
    @classmethod
    def _create_empty(cls, entity_id):
        """
        Creates a new instance with it's attribute set to their default values.
        
        Parameters
        ----------
        entity_id : `int`
            The entity's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.ended = False
        self.ends_at = None
        self.guild_id = 0
        self.id = entity_id
        self.paused_until = None
        self.user_id = 0
        return self
    

    @classmethod
    def precreate(cls, guild_boost_id, **keyword_parameters):
        """
        Creates a new guild boost instance. When the guild boost is loaded with the same id, it will be picked up.
        
        Parameters
        ----------
        guild_boost_id : `int`
            The guild boost's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        ended : `None | bool`, Optional (Keyword only)
            Whether the boost already ended.
        
        ends_at : `None | DateTime`, Optional (Keyword only)
            When the boost ends at.
        
        guild : ``None | int | Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        
        guild_id : ``None | int | Guild``, Optional (Keyword only)
            The guild's identifier the boost is for.
        
        paused_until : `None | DateTime`, Optional (Keyword only)
            Until when the boost is paused.
        
        user : ``None | int | ClientUserBase``, Optional (Keyword only)
            Alternative for `user_id`.
        
        user_id : ``None | int | ClientUserBase``, Optional (Keyword only)
            The user who owns the boost.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        guild_boost_id = validate_id(guild_boost_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = GUILD_BOOSTS[guild_boost_id]
        except KeyError:
            self = cls._create_empty(guild_boost_id)
            GUILD_BOOSTS[guild_boost_id] = self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    def copy(self):
        """
        Copies the guild boost.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.ended = False
        new.ends_at = None
        new.guild_id = 0
        new.id = 0
        new.paused_until = self.paused_until
        new.user_id = 0
        return new
    
    
    def copy_with(self, *, paused_until = ...):
        """
        Copies the guild boost with the given fields.
        
        Parameters
        ----------
        paused_until : `None | DateTime`, Optional (Keyword only)
            Until when the boost is paused.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # paused_until
        if paused_until is ...:
            paused_until = self.paused_until
        else:
            paused_until = validate_paused_until(paused_until)
        
        # Construct
        new = object.__new__(type(self))
        new.ended = False
        new.ends_at = None
        new.guild_id = 0
        new.id = 0
        new.paused_until = paused_until
        new.user_id = 0
        return new
    
    
    @property
    def guild(self):
        """
        Returns the guild boost's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the guild boost's user.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)
    
    
    @property
    def partial(self):
        """
        Returns whether the guild boost is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return False if self.id else True



__all__ = ()

from .constants import (
    EMBEDDED_ACTIVITY_UPDATE_CREATE, EMBEDDED_ACTIVITY_UPDATE_DELETE, EMBEDDED_ACTIVITY_UPDATE_USER_ADD,
    EMBEDDED_ACTIVITY_UPDATE_USER_DELETE
)
from .embedded_activity import EmbeddedActivity


def difference_handle_embedded_activity_update_event(data):
    """
    Handles embedded activity events returning the embedded activity and the changes with it.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Embedded activity update event.
    
    Returns
    -------
    embedded_activity : ``EmbeddedActivity``
        The updated embedded activity.
    
    changes : `list` of `tuple` (`int`, `object`)
        Change entries. Each tuple` has `2` elements:
        
        - The action identifier of the change:
        
            Can be one of the following:
            
            +---------------------------------------+-------+-----------+
            | Respective name                       | Value | Notes     |
            +=======================================+=======+===========+
            | EMBEDDED_ACTIVITY_UPDATE_NONE         | 0     |           |
            +---------------------------------------+-------+-----------+
            | EMBEDDED_ACTIVITY_UPDATE_CREATE       | 1     |           |
            +---------------------------------------+-------+-----------+
            | EMBEDDED_ACTIVITY_UPDATE_DELETE       | 2     |           |
            +---------------------------------------+-------+-----------+
            | EMBEDDED_ACTIVITY_UPDATE_UPDATE       | 3     | Not used. |
            +---------------------------------------+-------+-----------+
            | EMBEDDED_ACTIVITY_UPDATE_USER_ADD     | 4     |           |
            +---------------------------------------+-------+-----------+
            | EMBEDDED_ACTIVITY_UPDATE_USER_DELETE  | 5     |           |
            +---------------------------------------+-------+-----------+
        
        - The value in the context of the action.
            
            If `action` is `EMBEDDED_ACTIVITY_UPDATE_USER_ADD`, `EMBEDDED_ACTIVITY_UPDATE_USER_DELETE`, it will
            contain the joined or left users.
    """
    embedded_activity, created = EmbeddedActivity.from_data_is_created(data)
    
    changes = []
    
    # If the object was just created, we have 2 cases:
    # Just came to cache as a deleted activity / it was created
    if created:
        if embedded_activity.user_states:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_CREATE, None))
        else:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_DELETE, None))
    
    else:
        joined_users, left_users = embedded_activity._difference_update_user_states(data)
        for user in joined_users:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_USER_ADD, user))
        
        for user in left_users:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_USER_DELETE, user))
        
        if not embedded_activity.user_states:
            changes.append((EMBEDDED_ACTIVITY_UPDATE_DELETE, None))
    
    return embedded_activity, changes


def handle_embedded_activity_update_event(data):
    """
    Handles an embedded activity event.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Embedded activity update event.
    """
    embedded_activity, created = EmbeddedActivity.from_data_is_created(data)
    if not created:
        embedded_activity._update_user_states(data)



__all__ = ()

from ...field_parsers import bool_parser_factory, default_entity_parser_factory, nullable_string_parser_factory
from ...field_putters import (
    bool_optional_putter_factory, entity_putter_factory, nullable_string_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_validator_factory, nullable_string_validator_factory
)

from ..media_info import MediaInfo

from .constants import DESCRIPTION_LENGTH_MAX, DESCRIPTION_LENGTH_MIN


# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_optional_putter_factory('description')
validate_description = nullable_string_validator_factory(
    'description', DESCRIPTION_LENGTH_MIN, DESCRIPTION_LENGTH_MAX
)


# media

parse_media = default_entity_parser_factory('media', MediaInfo, default_factory = MediaInfo._create_empty)
put_media = entity_putter_factory('media', MediaInfo)


def validate_media(media):
    """
    Validates whether the given media is correct type.
    
    Parameters
    ----------
    media : ``str | MediaInfo``
        The media to validate.
    
    Returns
    -------
    media : ``MediaInfo``
    
    Raises
    ------
    TypeError
        - If `media` is given as an invalid type.
    """
    if isinstance(media, str):
        return MediaInfo(media)
    
    if isinstance(media, MediaInfo):
        return media
    
    raise TypeError(
        f'`media` can be `str | {MediaInfo.__name__}`, got {type(media).__name__}, {media!r}.'
    )


# spoiler

parse_spoiler = bool_parser_factory('spoiler', False)
put_spoiler = bool_optional_putter_factory('spoiler', False)
validate_spoiler = bool_validator_factory('spoiler', False)



__all__ = ('CONVERSION_NEST_MAIN',)

from ..constants import CONVERSION_KIND_FIELD
from ..conversion import Conversion


class CONVERSION_NEST_MAIN(Conversion):
    # Generic
    
    name = '__nest_main__'
    name_aliases = None
    expected_types_messages = 'tuple<str, dict<str, object>>'
    kind = CONVERSION_KIND_FIELD
    output_conversion = None
    
    # Setting
    
    set_merger = None
    set_type = None
    set_type_processor = None
    set_listing_identifier = None
    set_identifier = None
    
    def set_validator(value):
        # Omit validation.
        yield value
    
    # Reading
    
    get_default = None
    get_processor = None
    
    # Serialization
    
    serializer_key = None
    serializer_optional = None
    serializer_required = None

    def serializer_putter(main_data, required, value):
        key, sub_data = value
        if required or main_data:
            sub_data[key] = main_data
        return sub_data

    # Sorting
    sort_priority = 9000



__all__ = ()

from scarletio import copy_docs

from .base import ParameterConverterBase

try:
    # CPython
    from re import Pattern
except ImportError:
    # ChadPython (PyPy)
    from re import _pattern_type as Pattern


class ParameterConverterFormFieldKeyword(ParameterConverterBase):
    """
    Regex and string matcher and `custom_id` matching parameter parser for forms.
    
    Attributes
    ----------
    annotation : `str | re.Pattern`
        Annotation defaulting to the parameter's name if required.
    
    default : `object`
        Default value of the parameter.
    
    parameter_name : `str`
        The parameter's name.
    
    matcher : `FunctionType`
        Matches interaction options based on their `custom_id`.
    """
    __slots__ = ('annotation', 'default', 'matcher')
    
    def __new__(cls, parameter):
        """
        Creates a new parameter converter used by form submit fields.
        
        Parameters
        ----------
        parameter : ``Parameter``
            The parameter to create converter from.
        """
        # Default annotation to parameter name
        annotation = parameter.annotation
        if (annotation is None) or (not isinstance(annotation, (str, Pattern))):
            annotation = parameter.name
        
        if isinstance(annotation, str):
            matcher = cls._converter_string
        
        else:
            group_count = annotation.groups
            group_dict = annotation.groupindex
            group_dict_length = len(group_dict)
            
            if group_dict_length and (group_dict_length != group_count):
                raise ValueError(
                    f'Regex patterns with mixed dict groups and non-dict groups are disallowed, got '
                    f'{annotation!r}.'
                )
            
            if not group_count:
                matcher = cls._converter_regex
            
            elif group_dict_length:
                matcher = cls._converter_regex_group_dict
            
            else:
                matcher = cls._converter_regex_group_tuple
        
        self = object.__new__(cls)
        self.parameter_name = parameter.name
        self.annotation = annotation
        self.default = parameter.default
        self.matcher = matcher
        return self
    
    
    @copy_docs(ParameterConverterBase.__call__)
    async def __call__(self, client, interaction_event, value):
        return self.matcher(self, interaction_event)
    
    
    @copy_docs(ParameterConverterBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        # annotation
        repr_parts.append(', annotation = ')
        repr_parts.append(repr(self.annotation))
        
        # default
        repr_parts.append(', default = ')
        repr_parts.append(repr(self.default))
    
    
    @staticmethod
    def _converter_string(converter, interaction_event):
        """
        String form submit interaction option value matcher.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        value : `object | str | tuple<str>`
            The matched value or the converter's default value.
        """
        component_type, value_or_values = interaction_event.get_value_for(converter.annotation)
        
        resolve = component_type.resolve
        if (resolve is not None):
            value_or_values = resolve(interaction_event.resolved, value_or_values)
        
        if (value_or_values is None):
            value_or_values = converter.default
        
        return value_or_values
    
    
    @staticmethod
    def _converter_regex(converter, interaction_event):
        """
        Regex form submit interaction option value matcher.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        value : `object`
            The matched value or the converter's default value.
        """
        match, component_type, value_or_values = interaction_event.get_match_and_value(converter.annotation.fullmatch)
        
        resolve = component_type.resolve
        if (resolve is not None):
            value_or_values = resolve(interaction_event.resolved, value_or_values)
        
        if (value_or_values is None):
            value_or_values = converter.default
        
        return value_or_values
    
    
    @staticmethod
    def _converter_regex_group_dict(converter, interaction_event):
        """
        Regex form submit interaction option value matcher returning the matched group dictionary as well.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        groups : `None | dict<str, str>`
            The matched values by the regex pattern.
        
        value : `object | str | tuple<str>`
            The matched value or the converter's default value.
        """
        match, component_type, value_or_values = interaction_event.get_match_and_value(converter.annotation.fullmatch)
        
        resolve = component_type.resolve
        if (resolve is not None):
            value_or_values = resolve(interaction_event.resolved, value_or_values)
        
        if (value_or_values is None):
            value_or_values = converter.default
        
        if match is None:
            groups = None
        else:
            groups = match.groupdict()
        
        return groups, value_or_values
    
    
    @staticmethod
    def _converter_regex_group_tuple(converter, interaction_event):
        """
        Regex form submit interaction option value matcher returning the matched group tuple as well.
        
        Parameters
        ----------
        converter : ``ParameterConverterFormFieldKeyword``
            The parent converter instance using this function.
        
        interaction_event : ``InteractionEvent``
            A received interaction event.
        
        Returns
        -------
        groups : `None | tuple<str>`
            The matched values by the regex pattern.
        
        value : `object | str | tuple<str>`
            The matched value or the converter's default value.
        """
        match, component_type, value_or_values = interaction_event.get_match_and_value(converter.annotation.fullmatch)
        
        resolve = component_type.resolve
        if (resolve is not None):
            value_or_values = resolve(interaction_event.resolved, value_or_values)
        
        if (value_or_values is None):
            value_or_values = converter.default
        
        if match is None:
            groups = None
        else:
            groups = match.groups()
        
        return groups, value_or_values



__all__ = ()

from .fields import (
    put_description, put_enabled, put_welcome_channels, validate_description, validate_enabled,
    validate_welcome_channels
)


WELCOME_SCREEN_FIELD_CONVERTERS = {
    'description': (validate_description, put_description),
    'enabled': (validate_enabled, put_enabled),
    'welcome_channels': (validate_welcome_channels, put_welcome_channels),
}



__all__ = ('ComponentMetadataThumbnailMedia',)

from scarletio import copy_docs

from ..media_info import MediaInfo

from .base import ComponentMetadataBase
from .fields import (
    parse_description, parse_media, parse_spoiler, put_description, put_media, put_spoiler, validate_description,
    validate_media, validate_spoiler
)


class ComponentMetadataThumbnailMedia(ComponentMetadataBase):
    """
    Represents the metadata of a thumbnail media component.
    
    Attributes
    ----------
    description : `None | str`
        Description of the component's media.
    
    media : ``MediaInfo``
        The media of the component.
    
    spoiler : `bool`
        Whether the media should be spoilered.
    """
    __slots__ = ('description', 'media', 'spoiler')
    
    def __new__(cls, *, description = ..., media = ..., spoiler = ...):
        """
        Creates a thumbnail media component metadata.
        
        Parameters
        ----------
        description : `None | str`, Optional (Keyword only)
            Description of the component's media.
        
        media : ``str | MediaInfo``, Optional (Keyword only)
            The media of the component.
        
        spoiler : `bool` Optional (Keyword only)
            Whether the media should be spoilered.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # media
        if media is ...:
            media = MediaInfo._create_empty()
        else:
            media = validate_media(media)
        
        # spoiler
        if spoiler is ...:
            spoiler = False
        else:
            spoiler = validate_spoiler(spoiler)
        
        # Construct
        self = object.__new__(cls)
        self.description = description
        self.media = media
        self.spoiler = spoiler
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            description = keyword_parameters.pop('description', ...),
            media = keyword_parameters.pop('media', ...),
            spoiler = keyword_parameters.pop('spoiler', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # media
        repr_parts.append(' media = ')
        repr_parts.append(repr(self.media))
        
        # description
        description = self.description
        if (description is not None):
            repr_parts.append(', description = ')
            repr_parts.append(repr(description))
        
        # spoiler
        spoiler = self.spoiler
        if spoiler:
            repr_parts.append(', spoiler = ')
            repr_parts.append(repr(spoiler))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # media
        hash_value ^= hash(self.media)
        
        # spoiler
        hash_value ^= self.spoiler << 20
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # description
        if self.description != other.description:
            return False
        
        # media
        if self.media != other.media:
            return False
        
        # spoiler
        if self.spoiler != other.spoiler:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.description = parse_description(data)
        self.media = parse_media(data)
        self.spoiler = parse_spoiler(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = {}
        
        put_description(self.description, data, defaults)
        put_media(self.media, data, defaults, include_internals = include_internals)
        put_spoiler(self.spoiler, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.clean_copy)
    def clean_copy(self, guild = None):
        new = object.__new__(type(self))
        
        new.description = self.description
        new.media = self.media.copy()
        new.spoiler = self.spoiler
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.description = self.description
        new.media = self.media.copy()
        new.spoiler = self.spoiler
        
        return new
    
    
    def copy_with(
        self, 
        *,
        description = ...,
        media = ...,
        spoiler = ...,
    ):
        """
        Copies the thumbnail media component metadata with the given fields.
        
        Parameters
        ----------
        description : `None | str`, Optional (Keyword only)
            Description of the component's media.
        
        media : ``str | MediaInfo``, Optional (Keyword only)
            The media of the component.
        
        spoiler : `bool` Optional (Keyword only)
            Whether the media should be spoilered.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # media
        if media is ...:
            media = self.media.copy()
        else:
            media = validate_media(media)
        
        # spoiler
        if spoiler is ...:
            spoiler = self.spoiler
        else:
            spoiler = validate_spoiler(spoiler)
        
        # Construct
        new = object.__new__(type(self))
        new.description = description
        new.media = media
        new.spoiler = spoiler
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            description = keyword_parameters.pop('description', ...),
            media = keyword_parameters.pop('media', ...),
            spoiler = keyword_parameters.pop('spoiler', ...),
        )



from .array import *
from .constants import *
from .rtp_packet import *
from .utils import *
from .voice_packet import *


__all__ = (
    *array.__all__,
    *constants.__all__,
    *rtp_packet.__all__,
    *utils.__all__,
    *voice_packet.__all__,
)



__all__ = ()

from ...field_validators import force_date_time_validator_factory
from ...sticker import create_partial_sticker_data
from ...utils import DISCORD_EPOCH_START, datetime_to_timestamp, timestamp_to_datetime

from ..message import MessageFlag, MessageType
from ..message.fields import (
    parse_attachments as _parse_attachments, parse_components as _parse_components, parse_content as _parse_content,
    parse_edited_at as _parse_edited_at, parse_embeds as _parse_embeds, parse_flags as _parse_flags,
    parse_mentioned_role_ids as _parse_mentioned_role_ids, parse_mentioned_users as _parse_mentioned_users,
    parse_soundboard_sounds as _parse_soundboard_sounds, parse_stickers as _parse_stickers, parse_type as _parse_type,
    put_soundboard_sounds as _put_soundboard_sounds, validate_attachments, validate_components,
    validate_content, validate_edited_at, validate_embeds, validate_flags, validate_mentioned_role_ids,
    validate_mentioned_users, validate_soundboard_sounds, validate_stickers, validate_type
)


# attachments

def parse_attachments(data):
    """
    Parses out attachments value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    attachments : ``None | tuple<Attachment>``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_attachments(message_data)


def put_attachments(attachments, data, defaults):
    """
    Serializes the given attachments into the given data.
    
    Parameters
    ----------
    attachments : ``None | tuple<Attachment>``
        The attachments to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (attachments is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if attachments is None:
            attachment_datas = []
        else:
            attachment_datas = [
                attachment.to_data(defaults = defaults, include_internals = True) for attachment in attachments
            ]
        
        message_data['attachments'] = attachment_datas
    
    return data


# components

def parse_components(data):
    """
    Parses out components value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    components : ``None | tuple<Component>``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_components(message_data)


def put_components(components, data, defaults):
    """
    Serializes the given components into the given data.
    
    Parameters
    ----------
    components : ``None | tuple<Component>``
        The components to serialize.
    
    data : `dict<str, object>`
        Reaction event data.
    
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (components is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if components is None:
            component_datas = []
        else:
            component_datas = [
                component.to_data(defaults = defaults, include_internals = True)
                for component in components
            ]
        
        message_data['components'] = component_datas
    
    return data


# content

def parse_content(data):
    """
    Parses out content value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    content : `None | str`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_content(message_data)


def put_content(content, data, defaults):
    """
    Serializes the given content into the given data.
    
    Parameters
    ----------
    content : `None | str`
        The content to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (content is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if content is None:
            content = ''
        
        message_data['content'] = content
    
    return data


# created_at

def parse_created_at(data):
    """
    Parses out created at value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    created_at : `None | DateTime`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return DISCORD_EPOCH_START
    
    created_at = message_data.get('timestamp', None)
    if (created_at is None):
        return DISCORD_EPOCH_START
    
    return timestamp_to_datetime(created_at)


def put_created_at(created_at, data, defaults):
    """
    Serializes the given created at value into the given data.
    
    Parameters
    ----------
    created_at : `None | DateTime`
        The created at value to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (created_at != DISCORD_EPOCH_START) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        message_data['timestamp'] = datetime_to_timestamp(created_at)
    
    return data


validate_created_at = force_date_time_validator_factory('created_at')

# edited_at

def parse_edited_at(data):
    """
    Parses out edited at value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    edited_at : `None | DateTime`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_edited_at(message_data)


def put_edited_at(edited_at, data, defaults):
    """
    Serializes the given created at value into the given data.
    
    Parameters
    ----------
    edited_at : `None | DateTime`
        The created at value to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (edited_at is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if (edited_at is None):
            timestamp = None
        else:
            timestamp = datetime_to_timestamp(edited_at)
        
        message_data['edited_timestamp'] = timestamp
    
    return data


# embeds

def parse_embeds(data):
    """
    Parses out embeds value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    embeds : ``None | tuple<Embed>``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_embeds(message_data)


def put_embeds(embeds, data, defaults):
    """
    Serializes the given embeds into the given data.
    
    Parameters
    ----------
    embeds : ``None | tuple<Embed>``
        The embeds to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (embeds is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if embeds is None:
            embed_datas = []
        else:
            embed_datas = [embed.to_data(defaults = defaults) for embed in embeds]
        
        message_data['embeds'] = embed_datas
    
    return data


# flags

def parse_flags(data):
    """
    Parses out flags value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    flags : ``MessageFlag``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return MessageFlag(0)
    
    return _parse_flags(message_data)


def put_flags(flags, data, defaults):
    """
    Serializes the given flags into the given data.
    
    Parameters
    ----------
    flags : ``MessageFlag``
        The flags to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if flags or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        message_data['flags'] = int(flags)
    
    return data


# mentioned_users

def parse_mentioned_users(data, guild_id = 0):
    """
    Parses out mentioned users value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    message_mentioned_users : ``None | tuple<ClientUserBase>``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_mentioned_users(message_data, guild_id)


def put_mentioned_users(mentioned_users, data, defaults, *, guild_id = 0):
    """
    Serializes the given mentioned users into the given data.
    
    Parameters
    ----------
    mentioned_users : ``None | tuple<ClientUserBase>``
        The mentioned users to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's id where the message was created at.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (mentioned_users is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
    
        user_mention_datas = []
        
        if (mentioned_users is not None):
            for user in mentioned_users:
                user_data = user.to_data(defaults = defaults, include_internals = True)
                
                if guild_id:
                    try:
                        guild_profile = user.guild_profiles[guild_id]
                    except KeyError:
                        pass
                    else:
                        user_data['member'] = guild_profile.to_data(defaults = defaults, include_internals = True)
                
                user_mention_datas.append(user_data)
        
        message_data['mentions'] = user_mention_datas
    
    return data


# mentioned_role_ids

def parse_mentioned_role_ids(data):
    """
    Parses out mentioned role ids value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    mentioned_role_ids : ``MessageFlag``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_mentioned_role_ids(message_data)


def put_mentioned_role_ids(mentioned_role_ids, data, defaults):
    """
    Serializes the given mentioned_role_ids into the given data.
    
    Parameters
    ----------
    mentioned_role_ids : ``MessageFlag``
        The mentioned_role_ids to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if mentioned_role_ids or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if mentioned_role_ids is None:
            role_id_array = []
        else:
            role_id_array = [str(role_id) for role_id in mentioned_role_ids]
            
        message_data['mention_roles'] = role_id_array
    
    return data


# soundboard_sounds

def parse_soundboard_sounds(data):
    """
    Parses out soundboard_sounds value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    soundboard_sounds : ``None | tuple<SoundboardSound>``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_soundboard_sounds(message_data)


def put_soundboard_sounds(soundboard_sounds, data, defaults):
    """
    Serializes the given soundboard_sounds into the given data.
    
    Parameters
    ----------
    soundboard_sounds : ``None | tuple<SoundboardSound>``
        The soundboard_sounds to serialize.
    
    data : `dict<str, object>`
        Reaction event data.
    
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (soundboard_sounds is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        _put_soundboard_sounds(soundboard_sounds, message_data, defaults)
    
    return data


# stickers

def parse_stickers(data):
    """
    Parses out stickers value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    stickers : `None | tuple<Sticker>`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_stickers(message_data)


def put_stickers(stickers, data, defaults):
    """
    Serializes the given stickers into the given data.
    
    Parameters
    ----------
    stickers : `None | tuple<Sticker>`
        The stickers to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (stickers is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if stickers is None:
            sticker_datas = []
        else:
            sticker_datas = [create_partial_sticker_data(sticker) for sticker in stickers]
        
        message_data['sticker_items'] = sticker_datas
    
    return data


# type

def parse_type(data):
    """
    Parses out type value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    message_type : ``MessageType``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return MessageType.default
    
    return _parse_type(message_data)


def put_type(message_type, data, defaults):
    """
    Serializes the given type into the given data.
    
    Parameters
    ----------
    message_type : ``MessageType``
        The type to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (message_type is not MessageType.default) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        message_data['type'] = message_type.value
    
    return data



__all__ = ()


# attribute_name

def validate_attribute_name(attribute_name):
    """
    Validates the given audit log change attribute name.
    
    Parameters
    ----------
    attribute_name : `None | str`
        Attribute name to validate.
    
    Returns
    -------
    attribute_name : `str`
    
    Raises
    ------
    TypeError
        - `attribute_name` is not `None`, `str`.
    ValueError
        - `attribute_name` is not identifier.
    """
    if attribute_name is None:
        return ''
    
    if not isinstance(attribute_name, str):
        raise TypeError(
            f'`attribute_name` can be `None`, `str`, got {attribute_name.__class__.__name__}; {attribute_name!r}.'
        )
    
    if attribute_name and (not attribute_name.isidentifier()):
        raise ValueError(f'`attribute_name` can be an identifier, got {attribute_name!r}.')
    
    return attribute_name


# flags

def validate_flags(flags):
    """
    Validates the given audit log change flags.
    
    Parameters
    ----------
    flags : `None | int`
        Flags to validate.
    
    Returns
    -------
    flags : `int`
    
    Raises
    ------
    TypeError
        - `flags` is not `int`.
    ValueError
        - `flags` is negative.
    """
    if flags is None:
        return 0
    
    if not isinstance(flags, int):
        raise TypeError(f'`flags` can be `int`, got {flags.__class__.__name__}; {flags!r}.')
    
    if flags < 0:
        raise ValueError(f'`flags` can not be negative, got {flags!r}.')
    
    return flags



__all__ = ('GuildInventorySettings', )

from scarletio import RichAttributeErrorBaseType

from .fields import parse_emoji_pack_collectible, put_emoji_pack_collectible, validate_emoji_pack_collectible


class GuildInventorySettings(RichAttributeErrorBaseType):
    """
    A guild's inventory settings.
    
    Attributes
    ----------
    emoji_pack_collectible : `bool`
        Whether you can collect this guild's emojis and use it across all guilds.
    """
    __slots__ = ('emoji_pack_collectible', )
    
    def __new__(cls, *, emoji_pack_collectible = ...):
        """
        Creates a new guild inventory settings with the given fields.
        
        Parameters
        ----------
        emoji_pack_collectible : `bool`, Optional (Keyword only)
            Whether you can collect this guild's emojis and use it across all guilds.
        
        Raises
        ------
        TypeError
            - Parameter with incorrect type given.
        ValueError
            - Parameter with incorrect value given.
        """
        # emoji_pack_collectible
        if emoji_pack_collectible is ...:
            emoji_pack_collectible = False
        else:
            emoji_pack_collectible = validate_emoji_pack_collectible(emoji_pack_collectible)
        
        # Construct
        self = object.__new__(cls)
        self.emoji_pack_collectible = emoji_pack_collectible
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a guild inventory settings from the requested guild inventory settings data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Received guild inventory settings data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.emoji_pack_collectible = parse_emoji_pack_collectible(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the guild inventory settings to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_emoji_pack_collectible(self.emoji_pack_collectible, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the guild inventory settings' representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        emoji_pack_collectible = self.emoji_pack_collectible
        if emoji_pack_collectible:
            repr_parts.append(' emoji_pack_collectible = ')
            repr_parts.append(repr(emoji_pack_collectible))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two guild inventory settings are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.emoji_pack_collectible != other.emoji_pack_collectible:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the guild inventory settings."""
        hash_value = 0
        
        # emoji_pack_collectible
        hash_value ^= self.emoji_pack_collectible
        
        return hash_value
    
    
    def __bool__(self):
        """Returns whether the guild has any features."""
        if self.emoji_pack_collectible:
            return True
        
        return False
    
    
    def copy(self):
        """
        Copies the guild inventory settings.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.emoji_pack_collectible = self.emoji_pack_collectible
        return new
    
    
    def copy_with(self, *, emoji_pack_collectible = ...):
        """
        Copies the guild inventory settings with the given fields.
        
        Parameters
        ----------
        emoji_pack_collectible : `bool`, Optional (Keyword only)
            Whether you can collect this guild's emojis and use it across all guilds.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - Parameter with incorrect type given.
        ValueError
            - Parameter with incorrect value given.
        """
        # emoji_pack_collectible
        if emoji_pack_collectible is ...:
            emoji_pack_collectible = self.emoji_pack_collectible
        else:
            emoji_pack_collectible = validate_emoji_pack_collectible(emoji_pack_collectible)
        
        # Construct
        new = object.__new__(type(self))
        new.emoji_pack_collectible = emoji_pack_collectible
        return new



__all__ = ('Subscription',)

from ...bases import DiscordEntity
from ...core import SUBSCRIPTIONS
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import create_partial_user_from_id

from .fields import (
    parse_cancelled_at, parse_country_code, parse_current_period_end, parse_current_period_start, parse_entitlement_ids,
    parse_id, parse_renewal_sku_ids, parse_sku_ids, parse_status, parse_user_id, put_cancelled_at,
    put_country_code, put_current_period_end, put_current_period_start, put_entitlement_ids,
    put_id, put_renewal_sku_ids, put_sku_ids, put_status, put_user_id, validate_cancelled_at,
    validate_country_code, validate_current_period_end, validate_current_period_start, validate_entitlement_ids,
    validate_id, validate_renewal_sku_ids, validate_sku_ids, validate_status, validate_user_id
)
from .preinstanced import SubscriptionStatus


PRECREATE_FIELDS = {
    'cancelled_at': ('cancelled_at', validate_cancelled_at),
    'country_code': ('country_code', validate_country_code),
    'current_period_end': ('current_period_end', validate_current_period_end),
    'current_period_start': ('current_period_start', validate_current_period_start),
    'entitlement_ids': ('entitlement_ids', validate_entitlement_ids),
    'sku_ids': ('sku_ids', validate_sku_ids),
    'renewal_sku_ids': ('renewal_sku_ids', validate_renewal_sku_ids),
    'status': ('status', validate_status),
    'user': ('user_id', validate_user_id),
    'user_id': ('user_id', validate_user_id),
}


class Subscription(DiscordEntity):
    """
    Represents a subscription.
    
    Attributes
    ----------
    cancelled_at : `None | DateTime`
        Whether and when the subscription was cancelled.
    
    country_code : `None | str`
        `ISO3166-1` country code. Present when queried through oauth2.
    
    current_period_end : `None | DateTime`
        When the current subscription period ends.
    
    current_period_start : `None | DateTime`
        When the current subscription period started.
    
    entitlement_ids : `None | tuple<int>`
        The entitlements granted with this subscription.
    
    id : `int`
        The subscription's identifier.
    
    renewal_sku_ids : `None | tuple<int>`
        The third party stock keeping unit the subscription will be linked to after renewal.
    
    sku_ids : `None | tuple<int>`
        The third party stock keeping unit the subscription is linked to.
    
    status : ``SubscriptionStatus``
        The subscription's status.
    
    user_id : `int`
        The subscribed user's identifier.
    """
    __slots__ = (
        '__weakref__', 'cancelled_at', 'country_code', 'current_period_end', 'current_period_start', 'entitlement_ids',
        'renewal_sku_ids', 'sku_ids', 'status', 'user_id'
     )
    
    
    def __new__(cls, *, entitlement_ids = ..., renewal_sku_ids = ..., sku_ids = ..., user_id = ...):
        """
        Creates a new partial subscription.
        
        Parameters
        ----------
        entitlement_ids : `None | iterable<int> | iterable<Entitlement>`, Optional (Keyword only)
            The entitlements granted with this subscription.
        
        renewal_sku_ids : `None | iterable<int> | iterable<SKU>`, Optional (Keyword only)
            The third party stock keeping unit the subscription will be linked to after renewal.
        
        sku_ids : `None | iterable<int> | iterable<SKU>`, Optional (Keyword only)
            The third party stock keeping unit the subscription is linked to.
        
        user_id : `int | ClientUserBase`, Optional (Keyword only)
            The subscribed user's identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # entitlement_ids
        if entitlement_ids is ...:
            entitlement_ids = None
        else:
            entitlement_ids = validate_entitlement_ids(entitlement_ids)
        
        # renewal_sku_ids
        if renewal_sku_ids is ...:
            renewal_sku_ids = None
        else:
            renewal_sku_ids = validate_renewal_sku_ids(renewal_sku_ids)
        
        # sku_ids
        if sku_ids is ...:
            sku_ids = None
        else:
            sku_ids = validate_sku_ids(sku_ids)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self.cancelled_at = None
        self.country_code = None
        self.current_period_end = None
        self.current_period_start = None
        self.entitlement_ids = entitlement_ids
        self.id = 0
        self.renewal_sku_ids = renewal_sku_ids
        self.sku_ids = sku_ids
        self.status = SubscriptionStatus.active
        self.user_id = user_id
        return self
    
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new subscription from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Subscription data.
        
        Returns
        -------
        new : `instance<cls>`
        """
        subscription_id = parse_id(data)
        
        try:
            self = SUBSCRIPTIONS[subscription_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = subscription_id
            self._set_attributes(data)
            SUBSCRIPTIONS[subscription_id] = self
        else:
            self._set_attributes(data)
        
        return self
    
    
    @classmethod
    def from_data_is_created(cls, data):
        """
        Creates a new subscription 
        
        Parameters
        ----------
        data : `dict<str, object>`
            Subscription data.
        
        Returns
        -------
        new : `instance<cls>`
        created : `bool`
        """
        subscription_id = parse_id(data)
        
        try:
            self = SUBSCRIPTIONS[subscription_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = subscription_id
            self._set_attributes(data)
            SUBSCRIPTIONS[subscription_id] = self
            created = True
        else:
            self._set_attributes(data)
            created = False
        
        return self, created
    
    
    def _set_attributes(self, data):
        """
        Sets the subscription's attributes from the given data. (Except `.id`.)
        
        Parameters
        ----------
        data : `dict<str, object>`
            Subscription data.
        """
        self.entitlement_ids = parse_entitlement_ids(data)
        self.user_id = parse_user_id(data)
        
        self._update_attributes(data)
    

    
    def _update_attributes(self, data):
        """
        Updates the subscription by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Subscription data.
        """
        self.cancelled_at = parse_cancelled_at(data)
        self.country_code = parse_country_code(data)
        self.current_period_end = parse_current_period_end(data)
        self.current_period_start = parse_current_period_start(data)
        self.renewal_sku_ids = parse_renewal_sku_ids(data)
        self.sku_ids = parse_sku_ids(data)
        self.status = parse_status(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the attributes of the subscription and returns the changed ones within an `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Subscription data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            The updated attributes.
            
            The returned dictionary might contain the following items:
            
            +---------------------------+-----------------------------------------------+
            | Key                       | Value                                         |
            +===========================+===============================================+
            | cancelled_at              | `None | DateTime`                             |
            +---------------------------+-----------------------------------------------+
            | country_code              | `None`, `str`                                 |
            +---------------------------+-----------------------------------------------+
            | current_period_end        | `None | DateTime`                             |
            +---------------------------+-----------------------------------------------+
            | current_period_start      | `None | DateTime`                             |
            +---------------------------+-----------------------------------------------+
            | renewal_sku_ids           | `None | tuple<int>`                           |
            +---------------------------+-----------------------------------------------+
            | sku_ids                   | `None | tuple<int>`                           |
            +---------------------------+-----------------------------------------------+
            | status                    | ``SubscriptionStatus``                        |
            +---------------------------+-----------------------------------------------+
        """
        old_attributes = {}
        
        # cancelled_at
        cancelled_at = parse_cancelled_at(data)
        if self.cancelled_at != cancelled_at:
            old_attributes['cancelled_at'] = self.cancelled_at
            self.cancelled_at = cancelled_at
        
        # country_code
        country_code = parse_country_code(data)
        if self.country_code != country_code:
            old_attributes['country_code'] = self.country_code
            self.country_code = country_code
        
        # current_period_end
        current_period_end = parse_current_period_end(data)
        if self.current_period_end != current_period_end:
            old_attributes['current_period_end'] = self.current_period_end
            self.current_period_end = current_period_end
        
        # current_period_start
        current_period_start = parse_current_period_start(data)
        if self.current_period_start != current_period_start:
            old_attributes['current_period_start'] = self.current_period_start
            self.current_period_start = current_period_start
        
        # renewal_sku_ids
        renewal_sku_ids = parse_renewal_sku_ids(data)
        if self.renewal_sku_ids != renewal_sku_ids:
            old_attributes['renewal_sku_ids'] = self.renewal_sku_ids
            self.renewal_sku_ids = renewal_sku_ids
        
        # sku_ids
        sku_ids = parse_sku_ids(data)
        if self.sku_ids != sku_ids:
            old_attributes['sku_ids'] = self.sku_ids
            self.sku_ids = sku_ids
        
        # status
        status = parse_status(data)
        if self.status is not status:
            old_attributes['status'] = self.status
            self.status = status
        
        return old_attributes
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the subscription into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_cancelled_at(self.cancelled_at, data, defaults)
        put_country_code(self.country_code, data, defaults)
        put_current_period_end(self.current_period_end, data, defaults)
        put_current_period_start(self.current_period_start, data, defaults)
        put_entitlement_ids(self.entitlement_ids, data, defaults)
        put_id(self.id, data, defaults)
        put_renewal_sku_ids(self.renewal_sku_ids, data, defaults)
        put_sku_ids(self.sku_ids, data, defaults)
        put_status(self.status, data, defaults)
        put_user_id(self.user_id, data, defaults)
        
        return data
    
    
    @classmethod
    def _create_empty(cls, subscription_id):
        """
        Creates a new subscription instance with it's attribute set to their default values.
        
        Parameters
        ----------
        subscription_id : `int`
            The subscription's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.cancelled_at = None
        self.country_code = None
        self.current_period_end = None
        self.current_period_start = None
        self.entitlement_ids = None
        self.id = subscription_id
        self.renewal_sku_ids = None
        self.sku_ids = None
        self.status = SubscriptionStatus.active
        self.user_id = 0
        return self
    
    
    @classmethod
    def precreate(cls, subscription_id, **keyword_parameters):
        """
        Creates an subscription instance.
        
        Parameters
        ----------
        subscription_id : `int`
            The subscription's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        cancelled_at : `None | DateTime`, Optional (Keyword only)
            Whether and when the subscription was cancelled.
        
        country_code : `None | str`, Optional (Keyword only)
            `ISO3166-1` country code. Present when queried through oauth2.
        
        current_period_end : `None | DateTime`, Optional (Keyword only)
            When the current subscription period ends.
        
        current_period_start : `None | DateTime`, Optional (Keyword only)
            When the current subscription period started.
        
        entitlement_ids : `None | iterable<int> | iterable<Entitlement>`, Optional (Keyword only)
            The entitlements granted with this subscription.
        
        renewal_sku_ids : `None | iterable<int> | iterable<SKU>`, Optional (Keyword only)
            The third party stock keeping unit the subscription will be linked to after renewal.
        
        sku_ids : `None | iterable<int> | iterable<SKU>`, Optional (Keyword only)
            The third party stock keeping unit the subscription is linked to.
        
        status : `SubscriptionStatus | int`, Optional (Keyword only)
            The subscription's status.
        
        user : `int | ClientUserBase`, Optional (Keyword only)
            Alternative for `user_id`.
        
        user_id : `int | ClientUserBase`, Optional (Keyword only)
            The subscribed user's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        subscription_id = validate_id(subscription_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = SUBSCRIPTIONS[subscription_id]
        except KeyError:
            self = cls._create_empty(subscription_id)
            SUBSCRIPTIONS[subscription_id] = self
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    

    def __repr__(self):
        """Returns the subscription's representation."""
        repr_parts = ['<', type(self).__name__]
        
        subscription_id = self.id
        if subscription_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' user_id = ')
        repr_parts.append(repr(self.user_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two subscriptions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two subscriptions are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self.id != other.id:
                return False
        
        # cancelled_at -> ignore, internal
        # country_code -> ignore, internal
        # current_period_end -> ignore, internal
        # current_period_start -> ignore, internal
        
        # entitlement_ids
        if self.entitlement_ids != other.entitlement_ids:
            return False
        
        # renewal_sku_ids
        if self.renewal_sku_ids != other.renewal_sku_ids:
            return False
        
        # sku_ids
        if self.sku_ids != other.sku_ids:
            return False
        
        # status -> ignore, internal
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the hash value of the subscription."""
        subscription_id = self.id
        if subscription_id:
            return subscription_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the subscription's hash based on their fields.
        
        This method is called by ``.__hash__`` if the channel has no ``.id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # cancelled_at -> ignore, internal
        # country_code -> ignore, internal
        # current_period_end -> ignore, internal
        # current_period_start -> ignore, internal
        
        # entitlement_ids
        entitlement_ids = self.entitlement_ids
        if (entitlement_ids is not None):
            hash_value ^= len(entitlement_ids) << 4
            
            for entitlement_id in entitlement_ids:
                hash_value ^= entitlement_id
        
        # id -> ignore, internal
        
        # renewal_sku_ids
        renewal_sku_ids = self.renewal_sku_ids
        if (renewal_sku_ids is not None):
            hash_value ^= len(renewal_sku_ids)
            
            for renewal_sku_id in renewal_sku_ids:
                hash_value ^= renewal_sku_id
        
        # sku_ids
        sku_ids = self.sku_ids
        if (sku_ids is not None):
            hash_value ^= len(sku_ids) << 8
            
            for sku_id in sku_ids:
                hash_value ^= sku_id
        
        # status -> ignore, internal
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the subscription.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.cancelled_at = None
        new.country_code = None
        new.current_period_end = None
        new.current_period_start = None
        
        entitlement_ids = self.entitlement_ids
        if (entitlement_ids is not None):
            entitlement_ids = (*entitlement_ids,)
        new.entitlement_ids = entitlement_ids
          
        new.id = 0
        
        renewal_sku_ids = self.renewal_sku_ids
        if (renewal_sku_ids is not None):
            renewal_sku_ids = (*renewal_sku_ids,)
        new.renewal_sku_ids = renewal_sku_ids
        
        sku_ids = self.sku_ids
        if (sku_ids is not None):
            sku_ids = (*sku_ids,)
        new.sku_ids = sku_ids 
         
        new.status = SubscriptionStatus.active
        new.user_id = self.user_id
        
        return new
    
    
    def copy_with(self, *, entitlement_ids = ..., renewal_sku_ids = ..., sku_ids = ..., user_id = ...):
        """
        Copies the subscription with the given fields.
        
        Parameters
        ----------
        entitlement_ids : `None | iterable<int> | iterable<Entitlement>`, Optional (Keyword only)
            The entitlements granted with this subscription.
        
        renewal_sku_ids : `None | iterable<int> | iterable<SKU>`, Optional (Keyword only)
            The third party stock keeping unit the subscription will be linked to after renewal.
        
        sku_ids : `None | iterable<int> | iterable<SKU>`, Optional (Keyword only)
            The third party stock keeping unit the subscription is linked to.
        
        user_id : `int | ClientUserBase`, Optional (Keyword only)
            The subscribed user's identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # entitlement_ids
        if entitlement_ids is ...:
            entitlement_ids = self.entitlement_ids
            if (entitlement_ids is not None):
                entitlement_ids = (*entitlement_ids,)
        else:
            entitlement_ids = validate_entitlement_ids(entitlement_ids)
        
        # renewal_sku_ids
        if renewal_sku_ids is ...:
            renewal_sku_ids = self.renewal_sku_ids
            if (renewal_sku_ids is not None):
                renewal_sku_ids = (*renewal_sku_ids,)
        else:
            renewal_sku_ids = validate_renewal_sku_ids(renewal_sku_ids)
        
        # sku_ids
        if sku_ids is ...:
            sku_ids = self.sku_ids
            if (sku_ids is not None):
                sku_ids = (*sku_ids,)
        else:
            sku_ids = validate_sku_ids(sku_ids)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new.cancelled_at = None
        new.country_code = None
        new.current_period_end = None
        new.current_period_start = None
        new.entitlement_ids = entitlement_ids
        new.id = 0
        new.renewal_sku_ids = renewal_sku_ids
        new.sku_ids = sku_ids
        new.status = SubscriptionStatus.active
        new.user_id = user_id
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the entity is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (self.id == 0)
    
    
    @property
    def user(self):
        """
        The subscribed user.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)
    
    
    def iter_entitlement_ids(self):
        """
        Iterates over the entitlement identifiers of the subscription.
        
        This method is an iterable generator.
        
        Yields
        ------
        entitlement_id : `int`
        """
        entitlement_ids = self.entitlement_ids
        if (entitlement_ids is not None):
            yield from entitlement_ids
    
    
    def iter_renewal_sku_ids(self):
        """
        Iterates over the stock keeping unit identifiers of the subscription that it will be linked to after renewal.
        
        This method is an iterable generator.
        
        Yields
        ------
        renewal_sku_id : `int`
        """
        renewal_sku_ids = self.renewal_sku_ids
        if (renewal_sku_ids is not None):
            yield from renewal_sku_ids
    
    
    def iter_sku_ids(self):
        """
        Iterates over the stock keeping unit identifiers of the subscription.
        
        This method is an iterable generator.
        
        Yields
        ------
        sku_id : `int`
        """
        sku_ids = self.sku_ids
        if (sku_ids is not None):
            yield from sku_ids
    
    
    @property
    def sku_id(self):
        """
        Returns the first stock keeping unit identifier of the subscription.
        
        Returns
        -------
        sku_id : `int`
        """
        sku_ids = self.sku_ids
        if sku_ids:
            return sku_ids[0]
        
        return 0



__all__ = ('LIBRARY_USER_AGENT',)

from sys import implementation

from scarletio import IgnoreCaseMultiValueDictionary, IgnoreCaseString
from scarletio.web_common.headers import AUTHORIZATION, USER_AGENT

from ... import __version__
from ...env import API_VERSION, LIBRARY_AGENT_APPENDIX, LIBRARY_URL, LIBRARY_VERSION


# to receive
RATE_LIMIT_REMAINING = IgnoreCaseString('X-RateLimit-Remaining')
RATE_LIMIT_RESET = IgnoreCaseString('X-RateLimit-Reset')
RATE_LIMIT_RESET_AFTER = IgnoreCaseString('X-RateLimit-Reset-After')
RATE_LIMIT_LIMIT = IgnoreCaseString('X-RateLimit-Limit')
RATE_LIMIT_HASH = IgnoreCaseString('X-ratelimit-bucket')

# to send
AUDIT_LOG_REASON = IgnoreCaseString('X-Audit-Log-Reason')
RATE_LIMIT_PRECISION = IgnoreCaseString('X-RateLimit-Precision')
DEBUG_OPTIONS = IgnoreCaseString('X-Debug-Options')


# user agent
def build_user_agent(user_agent_base):
    """
    Builds the user agent header to use.
    
    Parameters
    ----------
    user_agent_base : `str`
        The base user agent to use.
    
    Returns
    -------
    user_agent : `str`
    """
    parts = []
    parts.append(user_agent_base)
    parts.append(' ')
    
    if LIBRARY_AGENT_APPENDIX is None:
        parts.append('Python (')
        parts.append(implementation.name)
        parts.append(' ')
        
        version = implementation.version
        parts.append(str(version[0]))
        parts.append('.')
        parts.append(str(version[1]))
        
        release_level = version[3]
        if release_level != 'final':
            parts.append(' ')
            parts.append(release_level)
        
        parts.append(')')
    
    else:
        parts.append(LIBRARY_AGENT_APPENDIX)
    
    return ''.join(parts)


# to send (generated)
if LIBRARY_VERSION is None:
    LIBRARY_VERSION = __version__

LIBRARY_USER_AGENT_BASE = f'DiscordBot ({LIBRARY_URL}, {LIBRARY_VERSION})'
LIBRARY_USER_AGENT = build_user_agent(LIBRARY_USER_AGENT_BASE)


# header building
def build_headers(bot, token, debug_options):
    """
    Builds the headers passed to discord.
    
    Parameters
    ----------
    bot : `bool`
        Whether the respective client is a bot.
    token : `str`
        The client's token.
    debug_options: `None | set<str>`
        Http debug options.
    
    Returns
    -------
    headers : ``IgnoreCaseMultiValueDictionary``
    """
    headers = IgnoreCaseMultiValueDictionary()
    headers[USER_AGENT] = LIBRARY_USER_AGENT
    headers[AUTHORIZATION] = f'Bot {token!s}' if bot else token
    
    if API_VERSION in (6, 7):
        headers[RATE_LIMIT_PRECISION] = 'millisecond'
    
    if (debug_options is not None):
        for debug_option in sorted(debug_options):
            headers[DEBUG_OPTIONS] = debug_option
    
    return headers



__all__ = ()

from base64 import b64decode as base_64_decode, b64encode as base_64_encode

from ...application import Application
from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, flag_parser_factory, float_parser_factory,
    force_string_parser_factory, int_parser_factory, nullable_date_time_parser_factory,
    nullable_entity_array_parser_factory, nullable_functional_parser_factory, nullable_string_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, flag_optional_putter_factory, float_optional_putter_factory,
    force_string_putter_factory, int_putter_factory, nullable_date_time_optional_putter_factory,
    nullable_entity_array_optional_putter_factory, nullable_string_optional_putter_factory,
    nullable_string_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, float_conditional_validator_factory,
    force_string_validator_factory, int_conditional_validator_factory, nullable_bytes_validator_factory,
    nullable_date_time_validator_factory, nullable_entity_array_validator_factory, nullable_entity_validator_factory,
    nullable_string_validator_factory, url_optional_validator_factory, url_required_validator_factory
)
from ...user import User

from .constants import DESCRIPTION_LENGTH_MAX, DURATION_DEFAULT
from .flags import AttachmentFlag


# application

parse_application = nullable_functional_parser_factory('application', Application.from_data_invite)


def put_application(application, data, defaults):
    """
    Serializes the attachment's application into the given data.
    
    Parameters
    ----------
    application : ``None | Application``
        The application to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (application is not None):
        if application is None:
            application_data = None
        else:
            application_data = application.to_data_invite(defaults = defaults, include_internals = True)
        
        data['application'] = application_data
    
    return data


validate_application = nullable_entity_validator_factory('application', Application)


# clip_created_at

parse_clip_created_at = nullable_date_time_parser_factory('clip_created_at')
put_clip_created_at = nullable_date_time_optional_putter_factory('clip_created_at')
validate_clip_created_at = nullable_date_time_validator_factory('clip_created_at')


# clip_users

parse_clip_users = nullable_entity_array_parser_factory('clip_participants', User)
put_clip_users = nullable_entity_array_optional_putter_factory(
    'clip_participants', User,force_include_internals = True
)
validate_clip_users = nullable_entity_array_validator_factory('clip_users', User)


# content_type

parse_content_type = nullable_string_parser_factory('content_type')
put_content_type = nullable_string_putter_factory('content_type')
validate_content_type = nullable_string_validator_factory('content_type', 0, 1024)

# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# duration

parse_duration = float_parser_factory('duration_secs', DURATION_DEFAULT)
put_duration = float_optional_putter_factory('duration_secs', DURATION_DEFAULT)
validate_duration = float_conditional_validator_factory(
    'duration',
    DURATION_DEFAULT,
    lambda duration : duration >= 0.0,
    '>= 0.0',
)

# flags

parse_flags = flag_parser_factory('flags', AttachmentFlag)
put_flags = flag_optional_putter_factory('flags', AttachmentFlag())
validate_flags = flag_validator_factory('flags', AttachmentFlag)

# height

parse_height = int_parser_factory('height', 0)
put_height = int_putter_factory('height')
validate_height = int_conditional_validator_factory(
    'height',
    0,
    lambda height : height >= 0,
    '>= 0',
)

# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id')

# name

parse_name = force_string_parser_factory('filename')
put_name = force_string_putter_factory('filename')
validate_name = force_string_validator_factory('name', 0, 1024)

# proxy_url

parse_proxy_url = nullable_string_parser_factory('proxy_url')
put_proxy_url = url_optional_putter_factory('proxy_url')
validate_proxy_url = url_optional_validator_factory('proxy_url')

# size

parse_size = int_parser_factory('size', 0)
put_size = int_putter_factory('size')
validate_size = int_conditional_validator_factory(
    'size',
    0,
    lambda size : size >= 0,
    '>= 0',
)

# temporary

parse_temporary = bool_parser_factory('ephemeral', False)
put_temporary = bool_optional_putter_factory('ephemeral', False)
validate_temporary = bool_validator_factory('temporary', False)

# title

parse_title = nullable_string_parser_factory('title')
put_title = nullable_string_optional_putter_factory('title')
validate_title = nullable_string_validator_factory('title', 0, 1024)

# url

parse_url = force_string_parser_factory('url')
put_url = url_optional_putter_factory('url')
validate_url = url_required_validator_factory('url')


# waveform

def parse_waveform(data):
    """
    Parses the waveform out from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    waveform : `None | bytes`
    """
    waveform = data.get('waveform')
    if (waveform is not None) and waveform:
        try:
            return base_64_decode(waveform)
        except ValueError:
            pass


def put_waveform(waveform, data, defaults):
    """
    Serializes the waveform into the given data.
    
    Parameters
    ----------
    waveform : `None | bytes`
        The waveform to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (waveform is not None) or defaults:
        if waveform is not None:
            waveform = base_64_encode(waveform).decode('ascii')
        
        data['waveform'] = waveform
    
    return data


validate_waveform = nullable_bytes_validator_factory('waveform', 0, 4096)


# width

parse_width = int_parser_factory('width', 0)
put_width = int_putter_factory('width')
validate_width = int_conditional_validator_factory(
    'width',
    0,
    lambda width : width >= 0,
    '>= 0',
)



__all__ = ()

from scarletio import RichAttributeErrorBaseType


class EventDeprecationDescriptor(RichAttributeErrorBaseType):
    """
    Descriptor added for deprecated events.
    
    Attributes
    ----------
    name : `str`
        The name of the deprecated event.
    
    deprecation : ``EventDeprecation``
        Deprecation notice.
    """
    __slots__ = ('deprecation', 'name')
    
    def __new__(cls, name, deprecation):
        """
        Creates a new event deprecation descriptor.
        
        Parameters
        ----------
        name : `str`
            The name of the deprecated event.
        
        deprecation : ``EventDeprecation``
            Deprecation notice.
        """
        self = object.__new__(cls)
        self.deprecation = deprecation
        self.name = name
        return self
    
    
    def __get__(self, instance, instance_type):
        """
        Gets the field with the conversion.
        If accessed as a type attribute returns itself.
        
        Parameters
        ----------
        instance : `None | EventHandlerPlugin`
            Instance to set the fields to.
        
        instance_type : ``EventHandlerPluginMeta``
            The instance's type.
        
        Returns
        -------
        value : `object | self`
        """
        if instance is None:
            return self
        
        deprecation = self.deprecation
        deprecation.trigger(self.name, 2)
        return getattr(instance, deprecation.use_instead)
    
    
    def __set__(self, instance, value):
        """
        Sets the field.
        
        Parameters
        ----------
        instance : `FlagBase`
            Instance to set the fields to.
        
        value : `object`
            The value to set.
        """
        deprecation = self.deprecation
        deprecation.trigger(self.name, 2)
        setattr(instance, deprecation.use_instead, value)

    
    def __delattr__(self, instance):
        """
        Deletes the field.
        
        Parameters
        ----------
        instance : `FlagBase`
            Instance to set the fields to.
        """
        deprecation = self.deprecation
        deprecation.trigger(self.name, 2)
        delattr(instance, deprecation.use_instead)
    
    
    def __repr__(self):
        """Returns the descriptor's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # name
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        # deprecation
        repr_parts.append(', deprecation = ')
        repr_parts.append(repr(self.deprecation))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two descriptors are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # deprecation
        if self.deprecation != other.deprecation:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the descriptor's hash value."""
        hash_value = 0
        
        # deprecation
        hash_value ^= hash(self.deprecation)
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value



__all__ = ('PollVoteDeleteEvent',)

from .add import PollVoteAddEvent


class PollVoteDeleteEvent(PollVoteAddEvent):
    """
    Represents a processed `MESSAGE_POLL_VOTE_REMOVE` dispatch event.
    
    Attributes
    ----------
    answer_id : `int`
        The voted answer's identifier.
    message : ``Message``
        The message from what the vote was removed.
    user_id : `int`
        The user's identifier who removed their vote.
    """
    __slots__ = ()



__all__ = ()

from ..audit_log_change.flags import FLAG_HAS_AFTER, FLAG_HAS_BEFORE


def change_serializer_modification(conversion, change):
    """
    Default serializer used for regular modification entries.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change : ``AuditLogChange``
        Audit log change to deserialize.
    
    Yields
    ------
    change_data : `dict<str, object>`
    """
    value_serializer = conversion.value_serializer
    flags = change.flags
    
    change_data = {
        'key': conversion.get_field_key(0),
    }
    
    if flags & FLAG_HAS_BEFORE:
        before = change.before
        if (value_serializer is None):
            value = before
        else:
            value = value_serializer(before)
        
        change_data['old_value'] = value
    
    if flags & FLAG_HAS_AFTER:
        after = change.after
        if (value_serializer is None):
            value = after
        else:
            value = value_serializer(after)
        
        change_data['new_value'] = value
    
    yield change_data


def change_serializer_addition_and_removal(conversion, change):
    """
    Serializer for additions and removals.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change : ``AuditLogChange``
        Audit log change to serialise.
    
    Yields
    ------
    change_data : `dict<str, object>`
    """
    value_serializer = conversion.value_serializer
    flags = change.flags
    
    if flags & FLAG_HAS_BEFORE:
        before = change.before
        if (value_serializer is None):
            value = before
        else:
            value = value_serializer(before)
        
        yield {
            'key': conversion.get_field_key(0),
            'new_value': value,
        }
    
    if flags & FLAG_HAS_AFTER:
        after = change.after
        if (value_serializer is None):
            value = after
        else:
            value = value_serializer(after)
        
        yield {
            'key': conversion.get_field_key(1),
            'new_value': value,
        }


def change_serializer_flattened_emoji(conversion, change):
    """
    Serializer for flattened emojis.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change : ``AuditLogChange``
        Audit log change to serialise.
    
    Yields
    ------
    change_data : `dict<str, object>`
    """
    flags = change.flags
    
    # Should not happen
    if not (flags & (FLAG_HAS_BEFORE | FLAG_HAS_AFTER)):
        return
    
    # Only before
    if flags & (FLAG_HAS_BEFORE | FLAG_HAS_AFTER) == FLAG_HAS_BEFORE:
        before = change.before
        if before is None:
            key = 'emoji_name'
            value = None
        elif before.is_unicode_emoji():
            key = 'emoji_name'
            value = before.unicode
        else:
            key = 'emoji_id'
            value = str(before.id)
        
        yield {
            'key': key,
            'old_value': value,
        }
        return
    
    # Only after
    if flags & (FLAG_HAS_BEFORE | FLAG_HAS_AFTER) == FLAG_HAS_AFTER:
        after = change.after
        if after is None:
            key = 'emoji_name'
            value = None
        elif after.is_unicode_emoji():
            key = 'emoji_name'
            value = after.unicode
        else:
            key = 'emoji_id'
            value = str(after.id)
        
        yield {
            'key': key,
            'new_value': value,
        }
        return
    
    
    after = change.after
    before = change.before
    
    # both before and after None
    if (before is None) and (after is None):
        # ????
        return
    
    # after not None
    if (before is None) and (after is not None):
        if after.is_unicode_emoji():
            key = 'emoji_name'
            value = None
        else:
            key = 'emoji_id'
            value = str(after.id)

        yield {
            'key': key,
            'old_value': None,
            'new_value': value,
        }
        return
    
    # before not None
    if (before is not None) and (after is None):
        if before.is_unicode_emoji():
            key = 'emoji_name'
            value = None
        else:
            key = 'emoji_id'
            value = str(before.id)

        yield {
            'key': key,
            'old_value': value,
            'new_value': None,
        }
        return
    
    # all not None
    # both unicode
    if before.is_unicode_emoji() and after.is_unicode_emoji():
        yield {
            'key': 'emoji_name',
            'old_value': before.unicode,
            'new_value': after.unicode,
        }
        return
    
    # both custom
    if before.is_custom_emoji() and after.is_custom_emoji():
        yield {
            'key': 'emoji_id',
            'old_value': str(before.id),
            'new_value': str(after.id),
        }
        return
    
    # mixed
    if before.is_unicode_emoji():
        key = 'emoji_name'
        value = before.unicode
    else:
        key = 'emoji_id'
        value = str(before.id)
    
    yield {
        'key': key,
        'old_value': value,
    }


    if after.is_unicode_emoji():
        key = 'emoji_name'
        value = after.unicode
    else:
        key = 'emoji_id'
        value = str(after.id)
    
    yield {
        'key': key,
        'new_value': value,
    }


def change_serializer_application_command_permission_overwrite(conversion, change):
    """
    Serializer for application command permission overwrites.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    conversion : ``AuditLogEntryChangeConversion``
        The conversion to pull its field of.
    change : ``AuditLogChange``
        Audit log change to serialise.
    
    Yields
    ------
    change_data : `dict<str, object>`
    """
    before = change.before
    after = change.after
    
    
    merged = {}
    
    if (before is not None):
        for before_entry in before:
            target_id = before_entry.target_id
            merged[target_id] = (before_entry, None)
    
    
    if (after is not None):
        for after_entry in after:
            target_id = after_entry.target_id
            
            entry = merged.get(target_id, None)
            if entry is None:
                before_entry = None
            else:
                before_entry = entry[0]
            
            merged[target_id] = (before_entry, after_entry)
    
    # Sort the items, so we can predetermine the order
    for target_id, (before_entry, after_entry) in sorted(merged.items()):
        change_data = {'key': str(target_id)}
        
        if (before_entry is not None):
            change_data['old_value'] = before_entry.to_data(defaults = True)
        
        if (after_entry is not None):
            change_data['new_value'] = after_entry.to_data(defaults = True)
        
        yield change_data



__all__ = ()

from ...discord.client import Client


UNLOADING_BEHAVIOUR_DELETE = 0
UNLOADING_BEHAVIOUR_KEEP = 1
UNLOADING_BEHAVIOUR_INHERIT = 2


SYNC_ID_GLOBAL = 0
SYNC_ID_MAIN = 1
SYNC_ID_NON_GLOBAL = 2


def raw_name_to_display(raw_name):
    """
    Converts the given raw application command name to it's display name.
    
    Parameters
    ----------
    raw_name : `str`
        The name to convert.
    
    Returns
    -------
    display_name : `str`
        The converted name.
    """
    return '-'.join([w for w in raw_name.strip('_ ').lower().replace(' ', '-').replace('_', '-').split('-') if w])


def normalize_description(description):
    """
    Normalizes a docstrings.
    
    Parameters
    ----------
    description : `None`, `str`
        The docstring to clear.
    
    Returns
    -------
    cleared : `None`, `str`
        The cleared docstring. If `docstring` was given as `None` or is detected as empty, will return `None`.
    """
    if description is None:
        return None
    
    lines = description.splitlines()
    
    # strip
    for index in range(len(lines)):
        lines[index] = lines[index].strip()
    
    # Remove empty lines from the start~
    while lines and not lines[0]:
        del lines[0]
    
    # Remove everything after the first empty line.
    try:
        pop_since_index = lines.index('')
    except ValueError:
        pop_since_index = -1
    
    if pop_since_index != -1:
        del lines[pop_since_index:]
    
    # Build response :KoishiSpam:
    if lines:
        return ' '.join(lines)


RUNTIME_SYNC_HOOKS = []

def runtime_sync_hook_is_client_running(client):
    """
    Runtime sync hook to check whether a slash command should be registered and synced instantly when added or removed.
    
    Parameters
    ----------
    client : ``Client``
        The respective client of the ``Slasher``.
    """
    return client.running

RUNTIME_SYNC_HOOKS.append(runtime_sync_hook_is_client_running)


